% Deconvolver32Channel.m Generation of IRs from EigenMike recorded sweep.
% This script takes sweep recorded files, deconvolve with inverse filters, trims IRs 
% and store the results in a .mat or .wav file 
%
% INPUT
% 32 ch sweep recorded .wav file from EigenMike
% inverse sweep
%
% OUTPUT
% IRs trimmed
% output file:		spsIRs.mat / spsIRs.wav
%
% Authors: Daniel Pinardi, Marco Binelli.
% Universitiy of Parma.
% 2017

clear all
close all
clc

%% TRIMMING PARAMETERS FOR IRs

irTargetLength          = 48000; % IR number of samples after cut
irPeakAnticipation      = 10000; % number of samples before IR peak (approximately)


%% LOAD DATA FROM FILES

addpath( './lib' )
configureFileNamesAndDirectories

% load 32ch EigenMike sweep recorded
disp( [ 'Loading 32 channels recorded sweep files: ' recordedSweepFile ] );
[ recordedSweep, fs ]   = audioread( recordedSweepFile );

% load inverse sweep 
disp( [ 'Loading inverse sweep file: ' invSweepFile ] );
[ inverseSweep, fs ]    = audioread( invSweepFile );


%% DECONVOLUTION
% Recorded sweep are convolved by inverse sweep giving IRs

numberChannels          = size ( recordedSweep, 2);                         
inverseSweepLength      = length( inverseSweep );
recordedSweepLength     = length( recordedSweep );
irLength                = recordedSweepLength + inverseSweepLength - 1;
irs                     = zeros( irLength, numberChannels );                % pre-allocate matrix for IRs

% Initialize string to show completed convolution percentage 
fprintf( '\nDeconvolving audio: ' )
percentageString        = '0%%';
percentageStringLenght  = length( percentageString );
fprintf( '%s', percentageString )
oldCompletedPercentage  = -1; 

for channelIndex = 1 : numberChannels
    
    % Updating string to show completed convolution percentage 
    CompletedPercentage = round( channelIndex / numberChannels * 100 );
    if oldCompletedPercentage ~= CompletedPercentage
        for i = 1:percentageStringLenght
            fprintf( '\b' )
        end
        percentageString        = sprintf( '%d%%', CompletedPercentage );
        percentageStringLenght  = length( percentageString );
        fprintf( '%s', percentageString )
    end
    oldCompletedPercentage      = CompletedPercentage;
    
    % perform deconvolution
    irs( :, channelIndex )      = fd_conv( recordedSweep( :, channelIndex ), inverseSweep ); 
end

fprintf( '\n' )

%% IR trim
% plot before trimming
scrsz                   = get( groot, 'ScreenSize' );
figure( 'Position', [ 1 1 scrsz(3) scrsz(4) ] )
subplot( 2, 1, 1 )
plot( irs );
title( 'Irs before trimming' );

[ ~ , peakIndex ]       = max(abs(irs(:,1)));                               % index of peak value of the first IR
iStart                  = peakIndex - irPeakAnticipation;                   % start index for IR trim
iStop                   = iStart + irTargetLength - 1;                      % stop index for IR trim
irsShort                = zeros( irTargetLength, numberChannels );          % pre-allocate matrix for trimmed IRs

for i = 1:numberChannels
    irsShort( :, i )    = irs( iStart : iStop, i );                         % perform trimming
    irsShort( :, i )    = irsShort( :, i ) - mean( irsShort( :, i ) );      % remove DC
end

% plot after trimming
subplot( 2, 1, 2 )
plot( irsShort );
title( 'Irs after trimming' );

fprintf( 'Check trimmed signal. Press any key to continue or CTRL+c to abort...\n' );
pause
close all

clear irs
irs                     = irsShort;                                         % rename
clear irsShort

%% Exporting results in .mat file

if ~isdir( irsFolder )
    mkdir( irsFolder )
end
disp( [ 'Saving 32 channels IRs MAT files to: ' irsMatFileName ] );
save( irsMatFileName, 'irs', 'fs' );


%% Normalize and save 32ch .wav file

maxAbsValue             = max( max( abs( irs ) ) );                         % search for absolute max value
irs                     = irs./ maxAbsValue;                                % normalize IRs between -1 and +1
disp( [ 'Saving 32 channels IRs WAV files to: ' irsWavFileName ] );
audiowrite( irsWavFileName, irs, fs, 'BitsPerSample', 32);

rmpath( './lib' )
fprintf( '\nEnded correctly.\n\n' );
