% GenerateSpsFormat.m Generation of SPS format from EigenMike recordings.
% This script takes 32 channels IRs or noise recording as input (both .mat 
% and .wav are accepted), convolves it with SPS matrix and stores the 
% results in a .mat or .wav file.
% 
%
% INPUT
% - IRs or noise file from EigenMike (.mat / .wav)
% - SPS matrix filter
%
% OUTPUT
% - SPS format (.mat / .wav)
%
%
% Authors: Daniel Pinardi, Marco Binelli.
% University of Parma.
% 2017

clear all
close all
clc

addpath( './lib' )
configureFileNamesAndDirectories

%% DEFINE SOURCE FILE TYPE. POSSIBLE VALUES 'irs' , 'noise'
% sourceFileType              = 'irs';
sourceFileType              = 'noise';


%% PARAMETER FOR MATRIX_CONVOLUTION
% if this dimension is exceed the function oa_multichannel_conv will be used
% depending on RAM of computer, might be necessary to decrease

limitLength                 = 250000;


%% LOAD DATA FROM FILES         ONLY one between .mat and .wav needed !

% IRs .mat file
if strcmp( sourceFileType, 'irs' )
    disp( [ 'IRS - Loading 32 channels files: ' irsMatFileName ] );
    load( irsMatFileName );
    fileToProcess           = irs;                                          % rename imported file
    clear irs
    outputWavFileName       = spsIrsWavFileName;
    outputMatFileName       = spsIrsMatFileName;
    if ~isdir( spsIrsFolder )
        mkdir( spsIrsFolder )
    end
end

% Noise .mat file
if strcmp( sourceFileType, 'noise' )
    disp( [ 'NOISE - Loading 32 channels files: ' recordedFileName ] );
    [ fileToProcess, fs ]   = audioread( recordedNoiseFile );
    outputWavFileName       = spsNoiseWavFileName;
    outputMatFileName       = spsNoiseMatFileName;
    if ~isdir( spsNoiseFolder )
        mkdir( spsNoiseFolder )
    end 
end
info                        = audioinfo( recordedNoiseFile );
bitsNumber                  = info.BitsPerSample;
fileToProcess               = single( fileToProcess );

% filter matrix for SPS beamforming
[ spsWav, fs ]              = audioread ( spsFilterMatrixFile );
virtualMicNumber            = size ( spsWav, 2 );
arraySize                   = size( fileToProcess, 2 );

fprintf( ['Loaded matrix for ' num2str(virtualMicNumber) ' directions.\n'] );
if arraySize == 32
    if virtualMicNumber == 32
        [ spsMatrix, fs ]       = load_ir( spsFilterMatrixFile, 32, 32 );   % spsMatrix is [32-IN x 32-OUT x Nsamples]
    else
        [ spsMatrix, fs ]       = load_ir( spsFilterMatrixFile, 32, 122 );  % spsMatrix is [32-IN x 122-OUT x Nsamples]
    end
elseif arraySize == 4
    if virtualMicNumber == 32
        [ spsMatrix, fs ]       = load_ir( spsFilterMatrixFile, 4, 32 );    % spsMatrix is [32-IN x 32-OUT x Nsamples]
    end
end

%% SPS transform

fileToProcessLength         = length( fileToProcess );

if  fileToProcessLength    <= limitLength
    % Reshaping matrix to use convolve library function: matrix_conv
    fileToProcessMatrix     = reshape( fileToProcess', 1, arraySize, fileToProcessLength );   
    spsFormat               = matrix_conv( fileToProcessMatrix, spsMatrix );        % convolving with SPS matrix
else
    % Reshaping matrix to use convolve library function: oa_multichannel_conv
    fileToProcessMatrix     = reshape( fileToProcess', arraySize, fileToProcessLength, 1 );   
    spsFormat               = oa_multichannel_conv( fileToProcessMatrix, spsMatrix );% convolving with SPS matrix
end
    
spsFormat                   = squeeze( spsFormat )';                        % eliminate exceeding matrix dimension
IR                          = mean( squeeze( spsMatrix( :, 1, : ) ) )';
[ ~, delay ]                = max( IR );
spsFormat                   = spsFormat( delay+1 : delay + length( fileToProcess ), : );


%% Exporting results in a .mat file

disp( [ 'Saving ' num2str( virtualMicNumber) ' channels SPS ' sourceFileType ' in MAT files: ' outputMatFileName ] );
save( outputMatFileName, 'spsFormat', 'fs', '-v7.3', '-nocompression' );


%% Normalize and save 32ch .wav file

maxAbsValue                 = max( max( abs( spsFormat ) ) );               % searc for absolute max value
spsFormat                   = spsFormat ./ maxAbsValue;                     % normalize IRs between -1 and +1

disp( [ 'Saving ' num2str( virtualMicNumber ) ' channels SPS ' sourceFileType ' in WAV files: ' outputWavFileName ] );
audiowrite( outputWavFileName, spsFormat, fs, 'BitsPerSample', bitsNumber );

rmpath( './lib' )
fprintf( '\nEnded correctly.\n\n' );
