% SpatialIRsMapper_32_122.m Spatial acoustic static mapping of an IR recording
% This script implements acoustic source mapping of an IR using 32ch or 
% 122ch recording in SPS format got from a spherical microphone array 
% (Eigenmike). It shows Matlab figures and saves png files containing pseudocolor SPL maps.
%
% INPUT
% - sps_format.mat -> file containing the SPS format for IR, generated by GenerateSpsFormat.m
% - panoramic_360_image.jpg
%
% OUTPUT
% - acousticMap.png
%
% Authors: Daniel Pinardi, Marco Binelli.
% University of Parma.
% 2017

clear all
close all
clc

addpath( './lib' )                                                          % add to path the library with needed common files
configureFileNamesAndDirectories                                            % configure filenames, directories, resources

set( 0, 'DefaultTextInterpreter', 'none' );                                 % for plot title visualization
scrsz                   = get( groot, 'ScreenSize' );                       % get screen size for plot
warning off                                                                 % to avoid warnings on image size

%% PARAMETERS
% EIGENMIKE -> GAIN = x  -  FULLSCALE = 112.9
% CYLINDER  -> GAIN = x  -  FULLSCALE = 106.3
% RICOH     -> GAIN = 0     FULLSCALE = 105.4
% AMBEO     -> GAIN = 0     FULLSCALE = 113.2
EM32gain                = 0;   % integer from 0 to 30
fullScaleLevel          = 112.9;                                            % dB

% pointing correction: + -> to the right / - -> to the left
horizontalTilt          = 0; % degrees

% image -  NB: aspect ratio is automatically setted to 2:1 !
imageOffsetCorrection   = 0;                                                % 1 -> enable
radiusCorrection        = 1.5;      % [m]
verticalOffset          = 0.2;      % [m] 
outputWindowWidth       = 2048;                                             % desired width of background image

% Trimming
trimEnable              = 0;                                                % enable trimming if 1
irPeakAnticipation      = 2500;                                             % number of samples before IR peak
irTargetLength          = 15000;                                            % IR number of samples after cut

% Signal processing
plotSignalEnable        = 0;
aWeightingEnable        = 1;                                                % enable to perform A filtering
nfft                    = 8192;                                             
overlap                 = 0.5;

% LOW PASS FILTER PARAMETERS
lowPassCutFreq          = 200;                                              % Cutoff Frequency [Hz]

% BAND PASS FILTER PARAMETERS
bandpassLowCutFreq      = 38;                                               % Low Cutoff Frequency [Hz]
bandpassHighCutFreq     = 44;                                               % High Cutoff Frequency [Hz]

% OCTAVE BAND FILTER PARAMETERS
% Rounded center freq are:
% [ 31 63 125 250 500 1000 2000 4000 8000 16000 ]
% [  1  2   3   4   5    6    7    8    9    10 ]
firstOctaveBand         = 1;                                                % select the first band to analyze
lastOctaveBand          = 1;                                                % select the last band to analyze

% Surf mapping
useFixedSplPlotBounds   = 0;                                                % if 0 each surface plot has color normalization 
fixedLowerLimit         = 80;                                               % user defined lower limit
fixedUpperLimit         = 90;                                               % user defined upper limit
noiseGate               = 4;                                               % max difference from maxlevel to plot
contourEnable           = 1;
valueTickEnable         = 0;
contourSpan             = 1;                                                % define contour line span
outputBlackWhite        = 1;                                                % 1 -> blackwhite, 0 -> color

% constant and other parameters
surfpoint               = 360;
alpha                   = .3;                                               % parameter for transparency mapping
correctionIndex         = round( horizontalTilt * surfpoint / 360 );


%% LOAD DATA FROM FILES
% define filename to load and folder output
mapsFolder              = mapsFolderIrs;
filename                = spsIrsMatFileName;
fprintf('\nIrs elaboration started\n');
 
% 32 or 122 SPS-irs
disp( ['Loading SPS format file: ' filename] );
load( filename );
virtualMicNumber        = size( spsIrs, 2 );                             % number of virtual microphones

% Background 360 image 
disp( ['Loading background image file: ' backgroundImageFile ] );
backgroundImage         = imread( backgroundImageFile );

%% IMAGE PROCESSING
% check image aspect ratio and perform correction if needed ( output format is 2x1 ) and resize image
[ backgroundImage, backgroundHeight, backgroundWidth, resizeFactor ] = image2x1resize( backgroundImage, outputWindowWidth );

% correction of offset between acoustic center and image center
if imageOffsetCorrection
    fprintf( 'Perform image correction\n\n' );
    [ backgroundImage ] = imageVertOffsetCorr( backgroundImage, radiusCorrection, verticalOffset );
end

if outputBlackWhite
    background          = (1+alpha) * flipud( backgroundImage( :, :, [ 1 1 1 ] ) ); % flip and conversion to black white
else
    background          = (1+alpha) * flipud( backgroundImage );                    % flip
end


%% A-weighting
if aWeightingEnable
    fprintf( 'Performing SPS format A-weighting.\n' );
    AWeighting          = weightingFilter( 'A-weighting', fs );
    for i = 1 : virtualMicNumber
        spsIrs( :, i )   = AWeighting( spsIrs( :, i ) );
    end
end


%% FULL SCALE LEVEL
fullScaleLevel      = 0.0004*( EM32gain^3 ) - 0.0078*( EM32gain^2 ) - 0.8821*( EM32gain ) + fullScaleLevel; % dB


%% TRIMMING
spsMean             = mean( spsIrs, 2 );                                 % mean signal for plot

if trimEnable
    
    fprintf( 'Perform trimming\n\n' );
    [ spsIrs ]      = irTrim( spsIrs, irPeakAnticipation, irTargetLength, plotSignalEnable );
    spsMean         = mean( spsIrs, 2 );                                 % update value
    
else % no trimming
    
    fprintf( 'Continuing without trimming.\n' );
    
    if plotSignalEnable
        plot( spsMean );
        title( 'Untrimmed signal' );
        fprintf( 'Check signal...then press any key to continue or Ctrl-C to abort\n\n' );
        pause
        close all
    end
end


%% FILTERING PROCESS
% Complex sum
[ complexSum, freqTot ] = computeComplexSum( spsIrs, [], [], [], virtualMicNumber, nfft, fs, overlap );
     
% Total unfiltered SPL
[ splTotal, limitFreq ] = computeSPLunfiltered( complexSum, fullScaleLevel, spsMean, fs, nfft, overlap, plotSignalEnable, [], [] );
 
% Lowpass filtered SPL
[ splLowPassFilt ]      = computeSPLlowpassfilt( complexSum, fullScaleLevel, freqTot, lowPassCutFreq, spsMean, ...
                            fs, nfft, overlap, plotSignalEnable, limitFreq, [], [] );

% Bandpass filtered SPL
[ splBandPassFilt ]     = computeSPLbandpassfilt( complexSum, fullScaleLevel, freqTot, bandpassLowCutFreq, bandpassHighCutFreq, ...
                            spsMean, fs, nfft, overlap, plotSignalEnable, limitFreq, [], [] );

% Octave band filtered SPL
[ splOctaveFilt, centerFreq ] = computeSPLoctavefilt( fs, firstOctaveBand, lastOctaveBand, plotSignalEnable, scrsz, ...
                                freqTot, complexSum, spsMean, nfft, overlap, limitFreq, fullScaleLevel, [], [] );

% Check plot if enabled
if plotSignalEnable
    fprintf( 'Check filtering...then press any key to continue or Ctrl-C to abort.\n\n' );
    pause
    close all
end
    
%% GENERATION OF SPL MATRIX
% first row             -> total value of SPL (without filtering)
% second row            -> total value of SPL (with low pass filtering)
% third row             -> total value of SPL (with band pass filtering)
% fourth to 'x' row     -> total value of SPL for each octave band computed

spl                     = [ splTotal; splLowPassFilt; splBandPassFilt; splOctaveFilt ];

% extend spl value to duplicated points (21 and 32) at top and bottom
if virtualMicNumber == 122
    [ spl ] = extendSPL122dir( spl );
    virtualMicNumberExtended = size( spl, 2 );
end

%% Build 122 or 32 points grid
if virtualMicNumber == 32
    fprintf( 'Building 32 virtual mics grid\n\n' );
    [ extX, extY, extXI, extYI, XI, YI, X, Y ] = GridBuilder32( backgroundWidth, backgroundHeight, surfpoint, virtualMicNumber );
else
    fprintf( 'Building 122 virtual mics grid\n\n' );
    [ extX, extY, extXI, extYI, XI, YI, X, Y ] = GridBuilder122( backgroundWidth, backgroundHeight, surfpoint, virtualMicNumber );
end

%% PLOT
numberPlot              = size(spl, 1);                                     % number of maps to be computed
refName                 = '';

% personalized colorscale
cmap                = jet( 64 );
cmap( 1, : )        = [ 0 0 0 ]; % black full scale

% show grid over background image and save image
figure( numberPlot + 1 );
gridOverImage( backgroundImage, virtualMicNumber, virtualMicNumberExtended, mapsFolder, X, Y );

% plot acoustic map
% 1                 -> total unfiltered SPL
% 2                 -> total SPL low pass filtered
% 3                 -> total SPL band pass filtered
% from 4 to end     -> SPL with octave filtering

for y = 1 : numberPlot
    splPlot             = spl( y, : );                                      % every cycle takes one row
    
    upperLimit          = ceil( max( max ( splPlot ) ) );
    lowerLimit          = upperLimit - noiseGate;

    [ ZI ]              = splInterpolation( double( splPlot ), virtualMicNumber, extX, extY, extXI, extYI, surfpoint );
    
    if correctionIndex > 0  
        ZI              = ZI( :, [ correctionIndex+1 : surfpoint, 1:correctionIndex ] );
    elseif correctionIndex < 0
        ZI              = ZI( :, [ surfpoint+correctionIndex+1 : surfpoint, 1:surfpoint+correctionIndex ] );
    end
            
    figure( y );
    
    % compute color map
    staticColorMap( background, backgroundWidth, backgroundHeight, XI, YI, ZI, lowerLimit, upperLimit, contourSpan, ...
                    alpha, contourEnable, useFixedSplPlotBounds, fixedLowerLimit, fixedUpperLimit, valueTickEnable, cmap );

    % print title for each map and save in png format, ordering by position
    saveColorMap( recordedFileName, spsFilterMatrixFileName, mapsFolder, lowPassCutFreq, ...
                  bandpassLowCutFreq, bandpassHighCutFreq, centerFreq, y, refName );
end

rmpath( './lib' )
fprintf( 'Ended correctly.\n' )
close all
 