% SpatialNoiseMapper_32_122.m Spatial acoustic static mapping of a NOISE recording
% This script implements acoustic source mapping of a noise recording
% using 32 ch 122ch recording in SPS format got from a spherical 
% microphone array (Eigenmike). It shows Matlab figures and saves png 
% files containing pseudocolor SPL maps.
%
% INPUT
% - sps_format.mat -> file containing the SPS format for NOISE, generated by GenerateSpsFormat.m
% - panoramic_360_image.jpg
%
% OUTPUT
% - acousticMap.png
%
% Authors: Daniel Pinardi, Marco Binelli.
% University of Parma.
% 2017

clear all
close all
clc

addpath( './lib' )                                                          % add to path the library with needed common files
configureFileNamesAndDirectories                                            % configure filenames, directories, resources

set( 0, 'DefaultTextInterpreter', 'none' );                                 % for plot title visualization
scrsz                   = get( groot, 'ScreenSize' );                       % get screen size for plot
warning off                                                                 % to avoid warnings on image size


%% PARAMETERS
% EIGENMIKE -> GAIN = x  -  FULLSCALE = 112.9
% CYLINDER  -> GAIN = x  -  FULLSCALE = 106.3
% RICOH     -> GAIN = 0     FULLSCALE = 105.4
% AMBEO     -> GAIN = 0     FULLSCALE = 113.2
EM32gain                = 0; % integer from 0 to 30
fullScaleLevel          = 112.9;                                              % dB

% pointing correction: + -> to the right / - -> to the left
horizontalTilt          = 0; % degrees

% COHERENT NOISE
coherenceEnable         = 0;                                                % if 1 -> performs coherent noise elaboration 
totalCoherenceMap       = 1;                                                % if 1 -> generate total coherent maps 
normalizeTotalCoherence = 1;
totalCoherenceNoiseGate = 3;     % [dB]

ref_01                  = 'errorMic_FL';    % errorMic_FL
ref_02                  = 'errorMic_FR';    % errorMic_FR
ref_03                  = '';               % puntone_FL
ref_04                  = '';               % puntone_FR
ref_05                  = '';               % punto_201
ref_06                  = '';               % center
ref_07                  = '';
ref_08                  = '';               % punto_452

% image -  NB: aspect ratio is automatically setted to 2:1 !
imageOffsetCorrection   = 0;                                                % 1 -> enable
radiusCorrection        = 1.5;      % [m]
verticalOffset          = 0.2;      % [m] 
outputWindowWidth       = 2048;                                             % desired width of background image

% Trimming
trimEnable              = 0;                                                % enable trimming if 1

% Signal processing
plotSignalEnable        = 1;
aWeightingEnable        = 1;                                                % enable to perform A filtering
nfft                    = 16384;                                            
overlap                 = 0.5;                                              

% LOW PASS FILTER PARAMETERS
lowPassCutFreq          = 1000;                                              % Cutoff Frequency [Hz]

% BAND PASS FILTER PARAMETERS
bandpassLowCutFreq      = 60;                                               % Low Cutoff Frequency [Hz]
bandpassHighCutFreq     = 630;                                               % High Cutoff Frequency [Hz]

% OCTAVE BAND FILTER PARAMETERS
% Rounded center freq are:
% [ 31 63 125 250 500 1000 2000 4000 8000 16000 ]
% [  1  2   3   4   5    6    7    8    9    10 ]
firstOctaveBand         = 1;                                                % select the first band to analyze
lastOctaveBand          = 1;                                                % select the last band to analyze

% Surf mapping
useFixedSplPlotBounds   = 0;                                                % if 0 each surface plot has color normalization 
fixedLowerLimit         = 80;                                               % user defined lower limit
fixedUpperLimit         = 90;                                               % user defined upper limit
noiseGate               = 10;                                               % max difference from maxlevel to plot
contourEnable           = 1;
valueTickEnable         = 1;
contourSpan             = 1;                                                % define contour line span
outputBlackWhite        = 1;                                                % 1 -> blackwhite, 0 -> color

% constant and other parameters
surfpoint               = 360;
alpha                   = 0.3;                                               % parameter for transparency mapping
correctionIndex         = round( horizontalTilt * surfpoint / 360 );


%% LOAD DATA FROM FILE
% define filename to load and folder output
refMicNumber            = 0;
referenceMic            = 0;
refName                 = '';
mimoEnable              = 0;

if coherenceEnable
    
    fprintf( '\nCoherent Noise elaboration started\n' );
    clear refMicNumber referenceMic refName
    mapsFolder          = mapsFolderNoise;
    filename            = spsNoiseMatFileName;
    
    fprintf( '\nLoading reference file\n' );
    [ refRec, fsReference ] = audioread( recordedReferenceSignalFile );
    
    % extract desired reference signal
    [ refMicNumber, referenceName, referenceMic ] = extractRef( refRec, ref_01, ref_02, ref_03, ref_04, ...
                                                    ref_05, ref_06, ref_07, ref_08 );
    
    if refMicNumber > 1
        mimoEnable = 1;
    end
    referenceMic        = single( referenceMic );
    clear refRec    

else
    mapsFolder          = mapsFolderNoise;
    filename            = spsNoiseMatFileName;
    fprintf('\nNoise elaboration started\n' );
end
    
% 32 or 122 SPS-noise
disp( ['Loading SPS format file: ' filename] );
load( filename );
spsFormat               = single( spsFormat );
virtualMicNumber        = size( spsFormat, 2 );                             % number of virtual microphones

% Background 360 image 
disp( ['Loading background image file: ' backgroundImageFile ] );
backgroundImage         = imread( backgroundImageFile );

%% IMAGE PROCESSING
% check image aspect ratio and perform correction if needed ( output format is 2x1 ) and resize image
[ backgroundImage, backgroundHeight, backgroundWidth, resizeFactor ] = image2x1resize( backgroundImage, outputWindowWidth );

% correction of offset between acoustic center and image center
if imageOffsetCorrection
    fprintf( 'Perform image correction\n\n' );
    [ backgroundImage ] = imageVertOffsetCorr( backgroundImage, radiusCorrection, verticalOffset );
end

if outputBlackWhite
    background          = (1+alpha) * flipud( backgroundImage( :, :, [ 1 1 1 ] ) ); % flip and conversion to black white
else
    background          = (1+alpha) * flipud( backgroundImage );                    % flip
end


%% A-weighting
if aWeightingEnable
    fprintf( 'Performing A-weighting.\n' );
    AWeighting          = weightingFilter( 'A-weighting', fs );
    for i = 1 : virtualMicNumber
        spsFormat( :, i ) = AWeighting( spsFormat( :, i ) );
    end
   
    if coherenceEnable
        for i = 1 : refMicNumber
            referenceMic( :, i ) = AWeighting( referenceMic( :, i ) );
        end
    end    
    fprintf( 'A-weighting ended.\n' );
end


%% FULL SCALE LEVEL
fullScaleLevel      = 0.0004*( EM32gain^3 ) - 0.0078*( EM32gain^2 ) - 0.8821*( EM32gain ) + fullScaleLevel; % dB


%% TRIMMING
% adjust spsFormat to reference length ( the difference cames from convolution with spsMatrix )
if coherenceEnable
    if length( spsFormat ) > length( referenceMic )
        spsFormat       = spsFormat( 1 : length( referenceMic ), : );
    else
        referenceMic    = referenceMic( 1 : length( spsFormat ), : );
    end
end

spsMean                 = mean( spsFormat, 2 );                             % mean signal for fft plot

if trimEnable
    
    fprintf( 'Perform trimming\n\n' );
    generateVideoJump   = 0;
    EMrecorded          = 0;
    [ spsFormat, referenceMic, EMrecorded ] = noiseTrim( spsFormat, spsMean, coherenceEnable, referenceMic, ...
                                              generateVideoJump, EMrecorded );
    
    spsMean             = mean( spsFormat, 2 );                             % update value

else % no trimming
    
    fprintf( 'Continuing without trimming.\n' );
    
    if plotSignalEnable
        plot( spsMean );
        title( 'Untrimmed signal' );
        fprintf( 'Check signal...then press any key to continue or Ctrl-C to abort\n\n' );
        pause
        close all
    end
    
end


%% FILTERING PROCESS
% Complex sum
[ complexSum, freqTot ] = computeComplexSum( spsFormat, coherenceEnable, refMicNumber, ...
                            referenceMic, virtualMicNumber, nfft, fs, overlap );
                       
% Total unfiltered SPL
[ splTotal, limitFreq ] = computeSPLunfiltered( complexSum, fullScaleLevel, spsMean, fs, nfft, overlap, ...
                            plotSignalEnable, coherenceEnable, refMicNumber );
     
% Lowpass filtered SPLG
[ splLowPassFilt ]      = computeSPLlowpassfilt( complexSum, fullScaleLevel, freqTot, lowPassCutFreq, spsMean, ...
                            fs, nfft, overlap, plotSignalEnable, limitFreq, coherenceEnable, refMicNumber );

% Bandpass filtered SPL
[ splBandPassFilt ]     = computeSPLbandpassfilt( complexSum, fullScaleLevel, freqTot, bandpassLowCutFreq, bandpassHighCutFreq, ...
                            spsMean, fs, nfft, overlap, plotSignalEnable, limitFreq, coherenceEnable, refMicNumber );

% Octave band filtered SPL
[ splOctaveFilt, centerFreq ] = computeSPLoctavefilt( fs, firstOctaveBand, lastOctaveBand, plotSignalEnable, scrsz, ...
                                freqTot, complexSum, spsMean, nfft, overlap, limitFreq, fullScaleLevel, coherenceEnable, refMicNumber );

% Check plot if enabled
if plotSignalEnable
    fprintf( 'Check filtering...then press any key to continue or Ctrl-C to abort.\n\n' );
    pause
    close all
end
    
%% GENERATION OF SPL MATRIX
% first row             -> total value of SPL (without filtering)
% second row            -> total value of SPL (with low pass filtering)
% third row             -> total value of SPL (with band pass filtering)
% fourth to 'x' row     -> total value of SPL for each octave band computed

if coherenceEnable
    for i = 1 : refMicNumber
        spl( :, :, i )  = [ splTotal( i, : ); splLowPassFilt( i, : ); splBandPassFilt( i, : ); splOctaveFilt( :, :, i ) ];
    end
else
    spl                 = [ splTotal; splLowPassFilt; splBandPassFilt; splOctaveFilt ];
end

% extend spl value to duplicated points (21 and 32) at top and bottom
if virtualMicNumber == 122
    [ spl ]             = extendSPL122dir( spl );
    virtualMicNumberExtended = size( spl, 2 );
else 
    virtualMicNumberExtended = 0;
end


%% Build 122 or 32 points grid
if virtualMicNumber == 32
    fprintf( 'Building 32 virtual mics grid\n\n' );
    [ extX, extY, extXI, extYI, XI, YI, X, Y ] = GridBuilder32( backgroundWidth, backgroundHeight, surfpoint, virtualMicNumber );
else
    fprintf( 'Building 122 virtual mics grid\n\n' );
    [ extX, extY, extXI, extYI, XI, YI, X, Y ] = GridBuilder122( backgroundWidth, backgroundHeight, surfpoint, virtualMicNumber );
end

%% PLOT
numberPlot              = size( spl, 1 );                                   % number of maps to be computed

% show grid over background image and save image
figure( numberPlot + 1 );
gridOverImage( backgroundImage, virtualMicNumber, virtualMicNumberExtended, mapsFolder, X, Y );

% personalized colorscale
cmap                = jet( 64 );
cmap( 1, : )        = [ 0 0 0 ]; % black full scale

% plot acoustic map
% 1                 -> total unfiltered SPL
% 2                 -> total SPL low pass filtered
% 3                 -> total SPL band pass filtered
% from 4 to end     -> SPL with octave filtering

if coherenceEnable
    
    for z = 1 : refMicNumber
        refName = deblank( referenceName( z, : ) );
        
        for y = 1 : numberPlot
            splPlot             = spl( y, :, z );                           % every cycle takes one row
            upperLimit          = ceil( max( max ( splPlot ) ) );
            lowerLimit          = upperLimit - noiseGate;

            [ ZI ]              = splInterpolation( double( splPlot ), virtualMicNumber, extX, extY, extXI, extYI, surfpoint );
            
            if correctionIndex > 0  
                ZI              = ZI( :, [ correctionIndex+1 : surfpoint, 1:correctionIndex ] );
            elseif correctionIndex < 0
                ZI              = ZI( :, [ surfpoint+correctionIndex+1 : surfpoint, 1:surfpoint+correctionIndex ] );
            end

            ZItot( :, :, y, z ) = ZI;
            
            figure( y );

            % compute color map
            staticColorMap( background, backgroundWidth, backgroundHeight, XI, YI, ZI, lowerLimit, upperLimit, contourSpan, ...
                            alpha, contourEnable, useFixedSplPlotBounds, fixedLowerLimit, fixedUpperLimit, valueTickEnable, cmap );

            % print title for each map and save in png format, ordering by position
            saveColorMap( recordedFileName, spsFilterMatrixFileName, mapsFolder, lowPassCutFreq, ...
                          bandpassLowCutFreq, bandpassHighCutFreq, centerFreq, y, refName );
        end
        close all
    end
    
else
    
    for y = 1 : numberPlot
        splPlot             = spl( y, : );                                      % every cycle takes one row
        upperLimit          = ceil( max( max ( splPlot ) ) );
        lowerLimit          = upperLimit - noiseGate;

        [ ZI ]              = splInterpolation( double( splPlot ), virtualMicNumber, extX, extY, extXI, extYI, surfpoint );
        
        if correctionIndex > 0  
            ZI              = ZI( :, [ correctionIndex+1 : surfpoint, 1:correctionIndex ] );
        elseif correctionIndex < 0
            ZI              = ZI( :, [ surfpoint+correctionIndex+1 : surfpoint, 1:surfpoint+correctionIndex ] );
        end
            
        figure( y );
        
        % compute color map
        staticColorMap( background, backgroundWidth, backgroundHeight, XI, YI, ZI, lowerLimit, upperLimit, contourSpan, ...
                        alpha, contourEnable, useFixedSplPlotBounds, fixedLowerLimit, fixedUpperLimit, valueTickEnable, cmap );

        % print title for each map and save in png format, ordering by position
        saveColorMap( recordedFileName, spsFilterMatrixFileName, mapsFolder, lowPassCutFreq, ...
                      bandpassLowCutFreq, bandpassHighCutFreq, centerFreq, y, refName )
    end
end

%% Total Coherence Map
if mimoEnable
    if totalCoherenceMap
        refName                         = 'totalCoherence';                 % update value

        % apply noise gate
        for i = 1 : numberPlot
            for j = 1 : refMicNumber
                maxValue( i, j )            = max( max( ZItot( :, :, i, j ) ) );
                totalCoherenceLimit( i, j ) = maxValue( i, j ) - totalCoherenceNoiseGate;
                for k = 1 : surfpoint/2
                    for y = 1 : surfpoint
                        if ZItot(k, y, i, j ) < totalCoherenceLimit( i, j )
                            ZItot( k, y, i, j ) = totalCoherenceLimit( i, j );
                        end
                    end
                end
            end
        end

        % normalizes between 0 and 1
        if normalizeTotalCoherence
            refName                     = 'totalCoherenceNormalized';       % update value
            contourSpan                 = 0.2;                              % update value
            useFixedSplPlotBounds       = 0;                                % update value
            for i = 1 : numberPlot
                for j = 1 : refMicNumber
                    ZItot( :, :, i, j ) = ZItot( :, :, i, j ) - totalCoherenceLimit( i, j );
                    maxValue( i, j )    = max( max( ZItot( :, :, i, j ) ) );
                    ZItot( :, :, i, j ) = ZItot( :, :, i, j ) ./ maxValue( i, j );
                end
            end
        end

        % define matrix of max values
        for i = 1 : numberPlot
            maxMatrix                   = max( ZItot( :, :, i, :) , [], 4 );
            maxMatrixType( :, :, i )    = maxMatrix;
        end

        for y = 1 : numberPlot
            ZI                  = maxMatrixType( :, :, y );
            upperLimit          = ceil( max( max ( ZI ) ) );
            lowerLimit          = floor( min( min( ZI) ) );

            figure( y );

            % compute color map
            staticColorMap( background, backgroundWidth, backgroundHeight, XI, YI, ZI, lowerLimit, upperLimit, contourSpan, ...
                            alpha, contourEnable, useFixedSplPlotBounds, fixedLowerLimit, fixedUpperLimit, valueTickEnable );

            % print title for each map and save in png format, ordering by position
            saveColorMap( recordedFileName, spsFilterMatrixFileName, mapsFolder, lowPassCutFreq, ...
                          bandpassLowCutFreq, bandpassHighCutFreq, centerFreq, y, refName );
        end
        close all  
    end
end

rmpath( './lib' )
fprintf( 'Ended correctly.\n' )
close all
 