% SpatialTimeIRsMapper_32_122.m Spatial acoustic video mapping of an IR recording
% This script implements source acoustic mapping of IRs using 32ch recordings 
% got from a spherical microphone array (Eigenmike).
% The interpolation method is the one developed by ing. Binelli
% Background is a static image.
%
% INPUT
% - sps_format.mat -> file containing the SPS format for IR, generated by GenerateSpsFormat.m
% - panoramic_360_image.jpg
%
% OUTPUT
% acusticIrVideo.mp4
%
% Authors: Daniel Pinardi, Marco Binelli.
% University of Parma.
% 2017

clear all
close all
clc

addpath( './lib' )                                                          % add to path the library with needed common files
configureFileNamesAndDirectories                                            % configure filenames, directories, resources

set( 0,'DefaultTextInterpreter', 'none' );                                  % setting for map titles
warning off                                                                 % to avoid warnings on image size


%% PARAMETERS
% Eigenmike gain
EM32gain                = 25; % EMIB gain, integer from 0 to 30 ( 0 for AMBEO and RICOH )
fullScaleLevel          = 112.9; % dB

% image -  NB: aspect ratio is automatically setted to 2:1 !
imageOffsetCorrection   = 0;
radiusCorrection        = 1.5;      % [m]
verticalOffset          = 0.2;      % [m] 
outputWindowWidth       = 1024;                                             % desired width of background image

% Trimming
trimEnable              = 1;                                                % enable trimming if 1
irPeakAnticipation      = 2500;                                             % number of samples before IR peak
irTargetLength          = 15000;                                            % IR number of samples after cut

% Signal processing
plotSignalEnable        = 0;
aWeightingEnable        = 1;                                                % enable to perform A filtering

% Filtering
filteringEnable         = 0;
bandpassLowCutFreq      = 50;                                               % Low Cutoff Frequency [Hz]
bandpassHighCutFreq     = 500;                                              % High Cutoff Frequency [Hz]
nfft                    = 4096;                                             % min value must be twice the difference
                                                                            % of filter window. If not, will be setted automatically.
% Video analysis
timeSignalEnable        = 1;                                                % plot time signal under video maps. Disable for VR
bufferSize              = 64;                                               % [samples]
stepSize                = 2;                                                % [samples]
frameRate               = 30;                                               % Default 30 

% !!!
% stepSize = (1-overlap)*bufferSize
% numframes = samples / stepSize
% ES: final length = 60s @ 30fps = 1800frames.
% If IR=2048 -> stepSize = 2048/1800 = 1.14
% Overap=0.975 -> bufferSize = 1.14/(1-0.975) = 45.6 samples

% Surf mapping
useFixedSplPlotBounds   = 0;                                                % if 0 each surface plot has color normalization 
fixedLowerLimit         = 80;                                               % user defined lower limit
fixedUpperLimit         = 90;                                               % user defined upper limit
noiseGate               = 10;                                               % max difference from maxlevel to plot
contourEnable           = 1;
valueTickEnable         = 0;
contourSpan             = 1;                                                % define contour line span
outputBlackWhite        = 1;                                                % 1 -> blackwhite, 0 -> color

% constant and other parameters
surfpoint               = 120;
alpha                   = 0.3;                                              % parameter for transparency mapping (0.4)
videoQuality            = 100;                                              % only for MPEG-4 format


%% LOAD DATA FROM FILES
% 32 or 122 SPS-irs
fprintf( '\nIrs elaboration started\n' );
filename                = spsIrsMatFileName;                                % define filename to load and folder output
disp( [ 'Loading SPS format files: ' filename ] );
load( filename );
virtualMicNumber        = size( spsIrs, 2 ); 

% Background 360 image 
disp( ['Loading background image file: ' backgroundImageFile ] );
backgroundImage         = imread( backgroundImageFile );


%% IMAGE PROCESSING
% check monitor dimension
% check monitor dimension and resize if necessary
[ outputWindowWidth ]   = screenSize( outputWindowWidth, timeSignalEnable );  

% check image aspect ratio and perform correction if needed ( output format is 2x1 ) and resize image
[ backgroundImage, backgroundHeight, backgroundWidth, resizeFactor ] = image2x1resize( backgroundImage, outputWindowWidth );

% correction of offset between acoustic center and image center
if imageOffsetCorrection
    fprintf( 'Perform image correction\n\n' );
    [ backgroundImage ] = imageVertOffsetCorr( backgroundImage, radiusCorrection, verticalOffset );
end

if outputBlackWhite
    background          = (1+alpha) * flipud( backgroundImage( :, :, [ 1 1 1 ] ) ); % flip and conversion to black white
else
    background          = (1+alpha) * flipud( backgroundImage );                    % flip
end


%% A-weighting
if aWeightingEnable
    fprintf( 'Performing SPS format A-weighting.\n' );
    AWeighting          = weightingFilter( 'A-weighting', fs );             % use Matlab built-in A_weigth filter
    for i = 1 : virtualMicNumber
        spsIrs( :, i ) = AWeighting( spsIrs( :, i ) );
    end
end


%% TRIMMING
if trimEnable
    
    fprintf( 'Perform trimming\n\n' );
    
    figure
    if aWeightingEnable
        plot( AWeighting( mean( spsIrs, 2 ) ) )
    else
        plot( mean( spsIrs, 2 ) )
    end
    title( 'Audio recorded' )

    prompt          = 'AUDIO samples to skip ( 0 is accepted ): ';
    noiseStart      = input( prompt );
    prompt          = 'last AUDIO sample: ';
    noiseEnd        = input( prompt );
    
    % update trimming value for video
    numbersFrame    = floor( ( noiseEnd - noiseStart ) / stepSize );        % integer number of frames
    
else % no trimming
   
    noiseStart      = 0;
    numbersFrame    = floor( length( spsIrs ) / stepSize );                        % integer number of frames

end

noiseEnd            = noiseStart + stepSize * numbersFrame;                 % update value
spsIrs              = spsIrs( 1 + noiseStart : noiseEnd, : );            % trim noise
close all

overlap             = 1 - stepSize / bufferSize;
fprintf( [ 'Overlap is ' num2str( overlap ) ' \n' ] )
fprintf( [ 'Frames are ' num2str( numbersFrame ) ' \n' ] )
fprintf( [ 'Video will be ' num2str( numbersFrame / frameRate ) 's \n' ] )


%% FULL SCALE LEVEL
fullScaleLevel      = 0.0004*( EM32gain^3 ) - 0.0078*( EM32gain^2 ) - 0.8821*( EM32gain ) + fullScaleLevel; % dB


%% PADDING
zeroPadMatrix       = zeros( bufferSize/2, virtualMicNumber );              % center of buffer must be at 0
spsIrs              = [ zeroPadMatrix; spsIrs; zeroPadMatrix ];          % apply zero padding
spsMean             = mean( spsIrs, 2 );                                 % update value


%% Build 122 or 32 points grid
if virtualMicNumber == 32
    fprintf( 'Building 32 virtual mics grid\n\n' );
    [ extX, extY, extXI, extYI, XI, YI, X, Y ] = GridBuilder32( backgroundWidth, backgroundHeight, surfpoint, virtualMicNumber );
else
    fprintf( 'Building 122 virtual mics grid\n\n' );
    [ extX, extY, extXI, extYI, XI, YI, X, Y ] = GridBuilder122( backgroundWidth, backgroundHeight, surfpoint, virtualMicNumber );
end

%% DATA PROCESSING
fprintf( 'Data processing started\n\n' );
if nfft < bufferSize
    nfft = bufferSize;
end

if filteringEnable
    % NFFT min value must be twice the difference of filter window
    if nfft < 2 * ( bandpassHighCutFreq - bandpassLowCutFreq )
        nfft = 2^nextpow2( 2 * ( bandpassHighCutFreq - bandpassLowCutFreq ) );
    end
    [ freqWindowing, idxLowBP, idxHighBP ] = bandpassFreqFilt( bandpassHighCutFreq, bandpassLowCutFreq, ...
                                                       nfft, fs, virtualMicNumber );
    freqWindowing = single( freqWindowing );
end
 
fftNormFactor       = sqrt( nfft * bufferSize );
splMatrix           = zeros ( surfpoint/2, surfpoint, numbersFrame );       % preallocation

% Percentage initializing
oldCompletedPercentage  = -1;
fprintf( '\nFrames processing: ' )
varStr                  = '0%%';
varStrLen               = length( varStr );
fprintf( '%s', varStr )

for blockIndex = 1 : numbersFrame
  
    %%%%%%%%%%% PERCENTAGE PROCESSING %%%%%%%%%%%  
    completedPercentage     = round( blockIndex/numbersFrame*100 );
    if oldCompletedPercentage ~= completedPercentage
        for i = 1 : varStrLen
            fprintf( '\b' )
        end
        varStr              = sprintf( '%d%%', completedPercentage );
        varStrLen           = length( varStr );
        fprintf( '%s', varStr )
    end
    oldCompletedPercentage  = completedPercentage;
    
    % perform FFT on a chunk and compute RMS with complex sum
    frfChunk     	= fft( spsIrs( (blockIndex-1) * stepSize + 1 : ...
                      (blockIndex-1) * stepSize + bufferSize, : ), nfft, 1 ) / fftNormFactor;
    if filteringEnable                   
        frf         = 2 * frfChunk( 1 : nfft/2 + 1, : ) .* freqWindowing;
    else
        frf         = 2 * frfChunk( 1 : nfft/2 + 1, : );
    end
    
    complexSum      = real( frf ).^2 + imag( frf ).^2;
    rmsChunk        = ( sum( complexSum, 1 ) )' ./ 2;
    splChunk        = ( 10 * log10( rmsChunk ) + fullScaleLevel )';
    
    if virtualMicNumber == 122
        [ splChunk ] = extendSPL122dir( splChunk ); % replicate SPL value on poles
    end
    
    % interpolation
    [ ZI ] = splInterpolation( double( splChunk ), virtualMicNumber, extX, extY, extXI, extYI, surfpoint );
    
    splMatrix( :, :, blockIndex )   = ZI;

end

upperLimit          = max( max( max( splMatrix ) ) );                       % find absolute max level
lowerLimit          = upperLimit - noiseGate;                               % range to plot

if useFixedSplPlotBounds == 1
    v               = fixedLowerLimit : contourSpan : fixedUpperLimit;      % iso-line levels
    splMatrix( splMatrix < fixedLowerLimit ) = 0;                           % set to 0 SPL values < lowerLimit
else                         
    v               = lowerLimit+1 : contourSpan : upperLimit;              % iso-line levels
    splMatrix( splMatrix < lowerLimit ) = 0;                                % set to 0 SPL values < lowerLimit
end

fprintf( 'Data processing ended correctly.\n\n' )

%% VIDEO PROCESSING
fprintf( 'Initializing video processing.\n\n' );
videoTitle          = recordedFileName( 1:end-4 );                          % name of file video
videoFile           = VideoWriter( videoTitle, 'MPEG-4' );                  % initializes video
videoFile.FrameRate = frameRate;                                            % set frame rate
videoFile.Quality   = videoQuality;                                         % set quality
open( videoFile );

cmap                = jet( 64 );
cmap( 1, : )        = [ 0 0 0 ]; % black full scale

% Percentage initializing
oldCompletedPercentage  = -1;
fprintf( '\nFrames processing: ' )
varStr                  = '0%%';
varStrLen               = length( varStr );
fprintf( '%s', varStr )

if timeSignalEnable
    
    videoHeigth     = outputWindowWidth / ( 2 * 0.8 );                      % define video window
    videoWindow     = figure( 'units', 'pixel', 'innerposition', [ 0 0 outputWindowWidth videoHeigth  ], 'menubar', 'none' );
    verticalLimit   = [ min( spsMean ) max( spsMean ) ];
    
    for blockIndex = 1 : numbersFrame
        
        %%%%%%%%%%% PERCENTAGE PROCESSING %%%%%%%%%%%  
        completedPercentage     = round( blockIndex/numbersFrame*100 );
        if oldCompletedPercentage ~= completedPercentage
            for i = 1 : varStrLen
                fprintf( '\b' )
            end
            varStr              = sprintf( '%d%%', completedPercentage );
            varStrLen           = length( varStr );
            fprintf( '%s', varStr )
        end
        oldCompletedPercentage  = completedPercentage;
    
        ZI          = splMatrix( :, :, blockIndex );                        % extract current value to plot
        
        % subplot colormap
        plotBG      = subplot( 'Position', [ 0 0.2 1 0.8 ] );
        videoColorMap( background, backgroundWidth, backgroundHeight, XI, YI, ZI, lowerLimit, upperLimit, plotBG , ...
                       v, alpha, contourEnable, useFixedSplPlotBounds, fixedLowerLimit, fixedUpperLimit, valueTickEnable, cmap );
        
        % subplot time signal
        plotTime   	= subplot( 'Position', [ 0 0 1 0.2 ] );           
        timeSignalPlot( spsMean, verticalLimit, blockIndex, stepSize, bufferSize );

        writeVideo( videoFile, getframe( videoWindow ) );

    end
    
else
    
    videoHeigth     = outputWindowWidth / 2;                                % set video window dimension to 2:1 for VR
    videoWindow   	= figure( 'units', 'pixel', 'innerposition', [ 0 0 outputWindowWidth videoHeigth  ], 'menubar', 'none' );

    for blockIndex = 1 : numbersFrame

        %%%%%%%%%%% PERCENTAGE PROCESSING %%%%%%%%%%%  
        completedPercentage     = round( blockIndex/numbersFrame*100 );
        if oldCompletedPercentage ~= completedPercentage
            for i = 1 : varStrLen
                fprintf( '\b' )
            end
            varStr              = sprintf( '%d%%', completedPercentage );
            varStrLen           = length( varStr );
            fprintf( '%s', varStr )
        end
        oldCompletedPercentage  = completedPercentage;
    
        ZI          = splMatrix( :, :, blockIndex );                        % extract current value to plot

        % subplot colormap
        plotBG      = subplot( 'Position', [ 0 0 1 1 ] );
        videoColorMap( background, backgroundWidth, backgroundHeight, XI, YI, ZI, lowerLimit, upperLimit, ...
                        plotBG, v, alpha, contourEnable, useFixedSplPlotBounds, fixedLowerLimit, fixedUpperLimit, valueTickEnable, cmap );
        
        writeVideo( videoFile, getframe( videoWindow ) );

    end
end

close( videoFile );                                                         % finalizes video 
rmpath( './lib' )
fprintf( '\n\nEnded correctly.\n\n' )
close all
