% SpatialTimeMapper122_chiesi.m Spatial acoustic mapping of a recording
% This script implements source acoustic mapping using 32ch recordings 
% (both noise and irs are supported) got from a spherical microphone array (Eigenmike).
% The interpolation method is the one developed by ing. Chiesi 
%
% INPUT
% - sps_format.mat -> file containing the SPS format for IR or NOISE, generated by
% GenerateSpsFormat.m
% - panoramic_360_image.jpg
%
% OUTPUT
% acusticVideo.mp4
%
% Authors: Lorenzo Chiesi.
% Universitiy of Parma.
% 2013 - revision 2017 by Daniel Pinardi


clear all
close all
clc

addpath( './lib' )
configureFileNamesAndDirectories

set( 0, 'DefaultTextInterpreter', 'none' );
warning off

%% PARAMETERS
% IRs or NOISE elaborarion
noise                   = 0 ;               % 1 -> noise ; 0 -> irs

% Trimming and normalization
aWeightingEnable        = 0;
normalizationEnable     = 1;
trimEnable              = 1;

% Image
blackwhiteEnable        = 1;            % set background image color ( 0 -> BW / 1 -> color )
outputWindowWidth       = 800;
imageOffsetCorrection   = 0;

% Video
onlyPanoEnable          = 0;            % if 1 enable only video mapping over background image (ir plot excluded)
frameRate               = 6; 
trasparency             = 0.4;

% Analysis parameter
virtualMicOrder         = 8;            % Has to be the same choosen when spsMatrix has been computed
bufferSize              = 64;          % samples
overlap                 = 0.75;
nfft                    = 4096;
lowCutFreq              = 100;
highCutFreq             = 1000;
dBMax                   = 0;
dBSpan                  = 30;
temperature             = 20;           % air temperature in Celsis


%% LOAD DATA FROM FILES
% define filename to load and folder output
if noise
    mapsFolder  = mapsFolderNoise;
    filename    = spsNoiseMatFileName;
    fprintf( '\nNoise elaboration started\n' );
else
    mapsFolder  = mapsFolderIrs;
    filename    = spsIrsMatFileName;
    fprintf( '\nIrs elaboration started\n' );
end

% 122 SPS file (ir or noise)
disp( [ 'Loading SPS format files: ' filename ] );
load( filename );
%spsFormat               = spsFormat';

% Background 360 image
disp( ['Loading background image file: ' backgroundImageFile] );
backgroundImage         = imread( backgroundImageFile );


%% VIRTUAL MICROPHONES directions and number
virtualMicNumber            = size( spsFormat, 2 );

if virtualMicNumber == 32
    
    GenerateEM32Dir;

    for i = 1 : 32
        if EM32Dir( i, 4 ) > 180 
           EM32Dir( i, 4 )  = EM32Dir( i, 4 ) - 360;
        end
    end

    pointsDegrees           = [ EM32Dir( :, 4 ), EM32Dir( :, 5 ) ];
    virtualMicDirections    = pointsDegrees / 180 * pi;
else
    virtualMicDirections    = GenerateUniformTestDir( 4 );
end


%% IMAGE PROCESSING
[ backgroundImage, backgroundHeight, backgroundWidth, resizeFactor ] = image2x1resize( backgroundImage, outputWindowWidth );
backgroundImage = double( backgroundImage ) / 255;

% correction of offset between acoustic center and image center
if imageOffsetCorrection
    fprintf( 'Perform image correction\n\n' );
    [ backgroundImage ] = imageVertOffsetCorr( backgroundImage, radiusCorrection, verticalOffset );
end

% convert to blackwhite if choosen
if blackwhiteEnable
    backgroundImage     = backgroundImage( :, :, [ 1 1 1 ] );
end


%% A-weighting
if aWeightingEnable
    fprintf( 'Performing SPS format A-weighting.\n' );
    for i = 1 : virtualMicNumber
        spsFormat( :, i ) = filterA( spsFormat( :, i ), fs );
    end
end


%% IR PROCESSING
stepSize                = bufferSize * ( 1 - overlap );

% normalization
if normalizationEnable
    spsFormat           = spsFormat / max( max( abs( spsFormat ) ) );
end

% Trimming
if trimEnable
    
    spsMean             = mean( spsFormat, 2 );                             % mean signal for plot
    figure
    plot( spsMean )
    title( 'Audio recorded by EigenMike' )

    prompt              = 'AUDIO samples to skip ( 0 is accepted ): ';
    noiseStart          = input( prompt );
    prompt              = 'last AUDIO sample: ';
    noiseEnd            = input( prompt );

    close all
    
    frameNumber         = floor( ( noiseEnd - noiseStart ) / stepSize );
    noiseEnd            = noiseStart + stepSize * frameNumber;              % update value
    spsFormat           = spsFormat( 1 + noiseStart : noiseEnd, : );        % trim
    
else % no trimming
    
    fprintf( 'Continuing without trimming.\n' );
    frameNumber         = floor( length( spsFormat ) / stepSize );
    spsFormat           = spsFormat( 1 : stepSize * frameNumber, : );       % trim
    
end

% Create omni IR and horizontal axis in meter to plot waveform
spsLength           = size( spsFormat, 1 );                                 % length of recorded irs after trim
omni_ir             = spsFormat( :, 1 )';
c                   = sqrt( 1.4 * 287.05 * ( temperature + 273.15 ) );
m_axis              = ( 0:1:spsLength-1 ) / fs * c;
m_axis_max          = max( m_axis );


%% COMPUTE FRAME ENERGY
timeWindow          = ones( 1, virtualMicNumber ) .* hann( bufferSize );    %Compute time domain analysis window
spectralLowLimit    = ceil( lowCutFreq / fs * nfft );                       %Compute spectral row lower limit
spectralHighLimit   = floor( highCutFreq / fs * nfft );                     %Compute spectral row higher limit
freqWindowing       = [ zeros( virtualMicNumber, spectralLowLimit ) ...
                        ones( virtualMicNumber, spectralHighLimit - spectralLowLimit + 1 ) ...
                        zeros( virtualMicNumber,( nfft / 2 ) - spectralHighLimit - 1 ) ]';
s                   = 1;
frameEnergy         = zeros( frameNumber, virtualMicNumber );

for f = 1 : 1 : frameNumber   % For each frame...
  
    % Compute complex pressure for each VMic
    complexPressure     = fft( spsFormat( s : s + bufferSize-1, : ) .* timeWindow, nfft, 1 );
    % Compute energy for each VMic
    frameEnergy( f, : ) = sum( ( abs( 2*complexPressure( 1:nfft/2, : ) ) .^2 ) .* freqWindowing , 1 );
    s                   = s + stepSize - 1;
    
end

% Global normalization 
frameEnergy             = frameEnergy / max( max( frameEnergy ) );


%% VIDEO PROCESSING
videoOverlay            = VideoWriter( 'AcousticOverlay', 'MPEG-4' );       % output name and format
videoOverlay.FrameRate  = frameRate;                                        % 6 is the minimum value for vlc reproduction
open( videoOverlay );

% Init color map interpolator
GenerateOverlayBase     = InitGenerateOverlay2( virtualMicDirections, virtualMicOrder, outputWindowWidth, backgroundHeight );

% Open figure to draw plots
h = figure( 'units', 'pixel', 'outerposition', [ 0 0 outputWindowWidth 2*backgroundHeight  ] );
s = 1;

if onlyPanoEnable
    
    for f = 1:1:frameNumber   % For each frame 
        
        % Elevare alla 1.5 migliora la definizione degli spot!
        EnergyImg       = GenerateOverlay2( GenerateOverlayBase, ( frameEnergy( f, : ) ) .^ ( 1.5 ), dBMax, dBSpan );
        PanoOverlayImg  = MixImage( EnergyImg, backgroundImage, trasparency );

        %% PLOT

        % Plot panoramic overlay image
        imshow( PanoOverlayImg )
        drawnow

        % Get figure as image and log to video file
        imgStruct       = getframe( h );
        writeVideo( videoOverlay, imgStruct.cdata )
        s               = s + stepSize - 1;

    end

else
    
    for f = 1:1:frameNumber   % For each frame 
            
        % Elevare alla 1.5 migliora la definizione degli spot!
        EnergyImg       = GenerateOverlay2( GenerateOverlayBase, ( frameEnergy(f,:)).^(1.5), dBMax, dBSpan );
        PanoOverlayImg  = MixImage( EnergyImg, backgroundImage, trasparency );

        %% PLOT

        % Plot panoramic overlay image
        subplot( 'Position', [ 0 0.5 1 0.5 ] )
        imshow( PanoOverlayImg )

        % Plot omni ir
        subplot( 'Position', [ 0.075 0.1 0.4 0.35 ] )
        frame_sel       = -ones( spsLength, 1 );
        frame_sel( s : s+bufferSize-1 ) = 1;
        plot( m_axis, omni_ir, m_axis, frame_sel, 'r' )
        axis( [ 0, m_axis_max, -1, +1 ] )

        % Plot omni ir detail
        subplot( 'Position', [ 0.525 0.1 0.4 0.35 ] )
        irplot( omni_ir( max( s - 0.5*bufferSize, 1 ) : s+1.5*bufferSize-1 ) )
        hold on 
        plot( [ zeros( 0.5*bufferSize, 1 ); hann( bufferSize ); zeros( 0.5*bufferSize, 1 ) ], 'r')
        hold off
        axis( [ 1, 2*bufferSize, -1, +1 ] )

        drawnow

        % Get figure as image and log to video file
        imgStruct       = getframe( h );
        writeVideo( videoOverlay, imgStruct.cdata )

        s               = s + stepSize - 1;

    end  
end

close( videoOverlay )
rmpath( './lib' )
fprintf( '\nEnded correctly\n' ) ;
