% SpatialTimeNoiseMapper_32_122_image.m Spatial acoustic mapping of a recording
% This script implements source acoustic mapping using 32ch recordings 
% of noise signal got from a spherical microphone array (Eigenmike).
% The interpolation method is the one developed by ing. Binelli
% Coherence elaboration is supported.
% Background is a static image.
%
%
% INPUT
% - sps_format.mat -> file containing the SPS format for NOISE, generated by GenerateSpsFormat.m
% - panoramic_360_image.jpg
%
% OUTPUT
% acusticVideo.mp4
%
% Authors: Daniel Pinardi, Marco Binelli.
% University of Parma.
% 2017-18

clear all
close all
clc

addpath( './lib' )                                                          % add to path the library with needed common files
configureFileNamesAndDirectories                                            % configure filenames, directories, resources
opengl( 'save', 'hardware' )												% set parameter for video processing
set( 0, 'DefaultTextInterpreter', 'none' );                                 % setting for map titles
warning off                                                                 % to avoid warnings on image size

%% PARAMETERS
% EIGENMIKE -> GAIN = x  -  FULLSCALE = 112.9
% CYLINDER  -> GAIN = x  -  FULLSCALE = 106.3
% RICOH     -> GAIN = 0     FULLSCALE = 105.4
% AMBEO     -> GAIN = 0     FULLSCALE = 113.2
EM32gain                = 25; % EMIB gain, integer from 0 to 30 ( 0 for AMBEO and RICOH )
fullScaleLevel          = 112.9; % dB

% pointing correction: + -> turn map to the right / - -> turn map to the left
horizontalTilt          = 0;        % degrees
surfpoint               = 120;      % number of horiz. interpolation points: increase up to 360 for horiz. correction

% coherence analysis
coherenceEnable         = 0;        % if 1 -> performs coherent noise elaboration 
totalTFenable           = 1;        % if 1, use only one transfer function calculated on whole signal. Good for time invariant system!
checkTransferFunction   = 0;        % compute TF with different averageWindow and ask to user the desired one

% reference signals to be used for coherence analysis
ref_01                  = '';
ref_02                  = '';
ref_03                  = '';
ref_04                  = '';
ref_05                  = '';
ref_06                  = '';
ref_07                  = '';
ref_08                  = '';

% image
imageOffsetCorrection   = 0;    	% 1 to enable. Correct vertical offset between acoustical and optical center
radiusCorrection        = 1.5;      % [m]. The radius of the sphere where correction is performed
verticalOffset          = 0.15;     % [m]. The amount of offset to be compensated 
mapWindowWidth          = 0;        % 0 to autoselect maximum possible, depending on screen resoution. Aspect ratio will be 2:1
outputBlackWhite        = 1;        % 1 -> blackwhite, 0 -> color
outputWidth             = 3840;     % 0 -> keep the same of colormap, xxxx -> rescale image and colormap
injectMetadata          = 0;        % 1 -> enable, 0 -> disable

% Signal processing
trimEnable              = 1;        % 1 to enable signal trimming
aWeightingEnable        = 1;        % 1 to enable A filtering

% Filtering
filteringEnable         = 0;       	% 1 to enable map filtering
filteringAudio          = 0;        % 1 to enable audio filtering - NB: with coherence analysis audio is filtered by the Transfer Function
bandpassLowCutFreq      = 50;       % Low Cutoff Frequency [Hz]
bandpassHighCutFreq     = 10000;    	% High Cutoff Frequency [Hz]

% Audio/Video analysis
ambixMatrixFileName     = 'Ambix_3order_2017.wav';
audioFromReference      = 0;        % 1 to enable autopanning reference with coherence analysis
audioFromSps            = 0;        % 1 to enable source selection from virtual mics
audioFromEm32           = 1;        % 1 to enable audio from Eigenmike recording
audioCoherence          = 0;        % 1 to enable coherent audio processing

audioMono               = 1;        % 1 to enable: add omni audio to .mp4
ambix_o1                = 0;        % 1 to enable: add ambix 1st order to map video
ambix_o3                = 0;        % 1 to enable: add ambix 3rd order to map video
audioBitsPerSample      = 32;       % 16, 24 or 32

timeSignalEnable        = 0;        % 1 to enable: plot time signal under video maps. Disable for VR
bufferSize              = 16384;    % [samples]
autoselectOverlap       = 1;        % enable for VR video
overlap                 = 0.975;    % user defined overlap                            
frameRate               = 30;       % Default 30 - min value for VLC is 6. 30 for VR

% Surf mapping
useFixedSplPlotBounds   = 0;        % if 0 each surface plot has color normalization with fixedLowerLimit and fixedUpperLimit
fixedLowerLimit         = 80;       % user defined lower limit
fixedUpperLimit         = 90;       % user defined upper limit
noiseGate               = 15;       % max difference from maxlevel to plot
contourEnable           = 1;        % 1 to enable isolevel lines
valueTickEnable         = 0;        % 1 to enable: show SPL value of isolevel lines
contourSpan             = 1;        % [dB] define isoline difference

% constant and other parameters
alpha                   = 0.3;      % parameter for transparency
videoQuality            = 100;      % quality of MPEG-4 format: decrease to reduce filesize


%% CHECK IF FFMPEG IS PRESENT IN C:\WINDOWS
if exist( 'C:\Windows\ffmpeg.exe', 'file' ) == 0
    error( 'FFMPEG not found. Please download from " https://ffmpeg.zeranoe.com/builds/ " and copy ffmpeg.exe in C:\Windows' );
end


%% LOAD DATA FROM FILES
% clear unused variables
clear backgroundVideoFile backgroundVideoFileName invSweepFile invSweepFileName irsFolder irsMatFileName irsWavFileName
clear mapsFolderIrs mapsFolderNoise outputFolder recordedNoiseFolder recordedReferenceSignalFileName recordedReferenceSignalFolder
clear recordedSweepFile recordedSweepFolder spsFilterMatrixFile spsFilterMatrixFileName spsIrsFolder spsIrsMatFileName
clear spsIrsWavFileName spsNoiseFolder spsNoiseWavFileName

% initialize variables
refMicNumber            = 0;
referenceMic            = 0;
mimoEnable              = 0;

% load reference files if necessary
if coherenceEnable || audioFromReference || audioCoherence
    
    fprintf( '\nCoherent Noise elaboration started\n' );
    clear refMicNumber referenceMic
    
    fprintf('\nLoading reference file\n' );
    [ refRec, fsReference ] = audioread( recordedReferenceSignalFile );     % read reference recorded files
    
    % extract desired reference signal, depending on previous selection
    [ refMicNumber, referenceName, referenceMic ] = extractRef( refRec, ref_01, ref_02, ref_03, ref_04, ...
                                                    ref_05, ref_06, ref_07, ref_08 );
    % enable MIMO analysis
    if refMicNumber > 1 
        referenceName   = 'multipleRef';
        mimoEnable      = 1;
        fprintf( '\nMIMO coherence enabled\n' )
    end
    referenceMic        = single( referenceMic );                           % convert data type 
    
else
    
    fprintf( '\nNoise elaboration started\n' );
    
end

% 32 or 122 SPS-noise
disp( [ 'Loading SPS format files: ' spsNoiseMatFileName ] );
load( spsNoiseMatFileName );
spsFormat               = single( spsFormat );                              % convert data type 
virtualMicNumber        = size( spsFormat, 2 );                             % define number of virtual mics: 32 or 122

% Background 360 image
disp( [ 'Loading background image file: ' backgroundImageFile ] );
backgroundImage         = imread( backgroundImageFile );

% CHECK AUDIO OPTION
if audioFromEm32 == 0 && audioFromSps == 0 && audioCoherence == 1  % audio coherence cannot be calculated
    error( 'ERROR: audio coherence cannot be calculated -> enable one between audioFromEm32 and audioFromSps' )
end

if audioFromEm32 && audioFromSps
    error( 'Enable only one between audioFromEm32, audioFromSps, audioFromReference' ) % only one possibility
elseif audioFromEm32 && audioFromReference
    error( 'Enable only one between audioFromEm32, audioFromSps, audioFromReference' ) % only one possibility
elseif audioFromSps && audioFromReference
    error( 'Enable only one between audioFromEm32, audioFromSps, audioFromReference' ) % only one possibility    
end

if audioMono == 0 && ambix_o1 == 0 && ambix_o3 == 0 % audio is not to be calculated: output will be mute.
    fprintf ( 'Output video will be mute.\n' )
    audioFromEm32       = 0;
    audioFromReference  = 0;
    audioFromSps      	= 0;
    audioCoherence      = 0;
elseif ( audioMono || ambix_o1 || ambix_o3 ) && (  audioFromEm32 == 0 && audioFromReference == 0 && audioFromSps == 0 )
    % audio has to be calculated: check one option is enable
    error ( 'ERROR: audio processing has to be perform -> check one option is enabled.\n' )
end


%% IMAGE PROCESSING
% check monitor dimension and resize if necessary
[ mapWindowWidth ]   = screenSize( mapWindowWidth, timeSignalEnable );                                                         % clear unused variables

% check image aspect ratio and perform correction if needed ( output format is 2x1 ) and resize image
[ backgroundImage, backgroundHeight, backgroundWidth, ~ ] = image2x1resize( backgroundImage, mapWindowWidth );

% correction of offset between acoustic center and image center
if imageOffsetCorrection
    fprintf( 'Perform image correction\n\n' );
    [ backgroundImage ]     = imageVertOffsetCorr( backgroundImage, radiusCorrection, verticalOffset );
end

if outputBlackWhite
    background              = (1+alpha)*flipud( backgroundImage( :, :, [ 1 1 1 ] ) ); % flip and conversion to black white
else
    background              = (1+alpha)*flipud( backgroundImage );                    % flip   
end
clear radiusCorrection verticalOffset


%% FULL SCALE LEVEL
fullScaleLevel      = 0.0004*( EM32gain^3 ) - 0.0078*( EM32gain^2 ) - 0.8821*( EM32gain ) + fullScaleLevel; % dB
clear EM32gain


%% Generate A-weighting filter
if aWeightingEnable
    fprintf( 'Generating A-weight filter.\n' );
    AWeighting          = weightingFilter( 'A-weighting', fs );             % use Matlab built-in A_weigth filter
end


%% Evaluate stepSize and overlap
if autoselectOverlap
    stepSize        = floor( fs / frameRate );
    overlap         = ( 1 - stepSize / bufferSize ) * 100;    
else
    stepSize        = floor( bufferSize * ( 1 - overlap ) );                % calculate stepSize
end
fprintf( [ '\nFrame rate is ', num2str( frameRate ) ' fps, fs is ' num2str( fs ) ' Hz and stepsize is ' num2str( stepSize ) ' samples. \n' ] );
fprintf( [ '\nBuffer size is ' num2str( bufferSize ) ' samples and video overlap is ', num2str( overlap ) '. \n' ] );


%% SYNC with REFERENCES
% synchronize reference signals ( if exist ) with SPS format
spsLength               = length( spsFormat );
delay                   = 0;
if numel( referenceMic ) ~= 1
    
    referenceLength     = length( referenceMic );
    % adjust reference to SPS length
    if referenceLength < spsLength
        referenceMic    = [ referenceMic; zeros( spsLength - referenceLength, size( referenceMic, 2 ) ) ];
    else
        referenceMic    = referenceMic( 1 : spsLength, : );
    end
    
    % estimate delay
    [ delay ]           = syncReference( spsFormat, referenceMic );

    % apply delay to references
    for i = 1 : refMicNumber
        if delay( i ) >= 0
            referenceMic( :, i )   = [ referenceMic( delay(i) + 1 : end, i ); zeros( delay(i), 1 ) ];
        else
            referenceMic( :, i )   = [ zeros( -delay(i), 1 ); referenceMic( 1 : end + delay(i), i ) ];
        end
    end    
end 


%% TRIM
if trimEnable
    
    figure
    if aWeightingEnable
        plot( AWeighting( mean( spsFormat, 2 ) ) )
    else
        plot( mean( spsFormat, 2 ) )
    end
    title( 'Audio recorded' )

    prompt          = 'AUDIO samples to skip ( 0 is accepted ): ';
    noiseStart      = input( prompt );
    prompt          = 'last AUDIO sample: ';
    noiseEnd        = input( prompt );

    close all

    numbersFrame    = floor( ( noiseEnd - noiseStart ) / stepSize );        % integer number of frames
else
    noiseStart      = 0;
    numbersFrame    = floor( spsLength / stepSize );                        % integer number of frames
end

noiseEnd            = noiseStart + stepSize * numbersFrame;                 % update value
spsFormat           = spsFormat( 1 + noiseStart : noiseEnd, : );            % trim noise

if numel( referenceMic ) ~= 1
    referenceMic    = referenceMic( 1 + noiseStart : noiseEnd, : );         % trim reference
end


%% ZERO PADDING
zeroPadMatrix       = zeros( bufferSize/2, virtualMicNumber );              % center of buffer must be at 0
spsFormat           = [ zeroPadMatrix; spsFormat; zeroPadMatrix ];          % apply zero padding

if numel( referenceMic ) ~= 1
    referenceMic    = [ zeroPadMatrix( :, 1:refMicNumber ); referenceMic; zeroPadMatrix( :, 1:refMicNumber ) ]; % apply zero padding
    referenceLength = length( referenceMic );                               % update value
end


%% PERFORM A-WEIGHTING
if aWeightingEnable
    fprintf( 'A-weight filtering started.\n' );
    if audioFromSps
        spsAudio            = spsFormat;    % duplicate for avoiding A-filtering
    end
    
    if audioFromReference || audioCoherence
        referenceMicAudio   = referenceMic; % duplicate for avoiding A-filtering
    end
    
    for i = 1 : virtualMicNumber
        spsFormat( :, i )   = AWeighting( spsFormat( :, i ) );
    end
   
    if coherenceEnable
        for i = 1 : refMicNumber
            referenceMic( :, i ) = AWeighting( referenceMic( :, i ) );
        end
    end
        
    fprintf( 'A-weighting ended.\n' );
end


%% Build 122 or 32 points grid
if virtualMicNumber == 32
    fprintf( '\nBuilding 32 virtual mics grid\n' );
    [ extX, extY, extXI, extYI, XI, YI, X, Y ] = GridBuilder32( backgroundWidth, backgroundHeight, surfpoint, virtualMicNumber );
else
    fprintf( '\nBuilding 122 virtual mics grid\n' );
    [ extX, extY, extXI, extYI, XI, YI, X, Y ] = GridBuilder122( backgroundWidth, backgroundHeight, surfpoint, virtualMicNumber );
end


%% DATA PROCESSING
% initialize variables
freqWindowing       = 0;
transferFunction    = 0;
fftStepNumber       = 0;
fftStep             = 0;
win                 = 0;
energyNormFactor    = 0;

if coherenceEnable && checkTransferFunction % define nfft value
    if totalTFenable
        fprintf( [ '\nbufferSize is ' num2str( bufferSize ) ' samples. Nfft must be = or < \n' ] )
        [ nfft ] = checkTFnfft( referenceMic, spsFormat );
        if nfft > bufferSize
            fprintf( [ 'Nfft ( ' num2str( nfft ) ' ) > bufferSize ( ' num2str( bufferSize ) ' ) is not allowed.\n' ] )
            fprintf( 'Type CTRL+C to abort and change bufferSize value or ...\n' )
            prompt                  = '...type nfft value: ';
            nfft                    = input( prompt );
        end
    else
        fprintf( [ 'bufferSize is ' num2str( bufferSize ) ' samples. Nfft must be = or < \n' ] )
        [ nfft ] = checkTFnfft2( referenceMic, spsFormat );
        if nfft > bufferSize
            fprintf( [ 'Nfft ( ' num2str( nfft ) ' ) > bufferSize ( ' num2str( bufferSize ) ' ) is not allowed.\n' ] )
            fprintf( 'Type CTRL+C to abort and change bufferSize value or ...\n' )
            prompt                  = '...type nfft value: ';
            nfft                    = input( prompt );
        end
    end
else
    fprintf( [ 'NFFT value setted equal to bufferSize: ' num2str( bufferSize ) '\n'] )
    nfft            = bufferSize;
    fprintf( 'Press any key to continue or CTRL+C to abort\n' )
    pause
end

if coherenceEnable && totalTFenable == 0
    % check nfft: must be =< bufferSize
    if nfft > bufferSize
        fprintf( [ 'NFFT ( ' num2str( nfft ) ' ) is too large for computing transfer function each frame\n' ] )
        fprintf( [ 'NFFT has been setted equal to the max allowed value, that is bufferSize ( ' num2str( bufferSize ) ' )\n' ] );
        fprintf( 'Press any key to continue or CTRL+C to abort\n' )
        pause
    end
end

% check nfft, update value for correct filtering and generate proper freq. window
if filteringEnable
    if nfft < 2 * ( bandpassHighCutFreq - bandpassLowCutFreq )
        fprintf( [ '\nFilter window ' num2str( bandpassLowCutFreq ) ' - ' num2str( bandpassHighCutFreq ) ' Hz is too large\n' ] );
        fprintf( [ 'NFFT is ' num2str( nfft ) '\n' ] )
        bandpassHighCutFreq = bandpassLowCutFreq + nfft/2;  % update vale to max possible
        fprintf( [ 'Filter window has been setted to ' num2str( bandpassLowCutFreq ) ' - ' num2str( bandpassHighCutFreq ) ' Hz\n' ] );
        fprintf( 'Tap a key to continue or CTRL+C to stop\n' )
        pause
    end
    
    [ freqWindowing, idxLowBP, idxHighBP ] = bandpassFreqFilt( bandpassHighCutFreq, bandpassLowCutFreq, ...
                                                       nfft, fs, virtualMicNumber ); % generate freq. window
    freqWindowing      = single( freqWindowing ); % convert data type
         
    if mimoEnable % update freq. window for MIMO analysis
             
        clear freqWindowing
        freqWindowing  = [ zeros( idxLowBP, refMicNumber ); ...
                           ones( idxHighBP - idxLowBP + 1, refMicNumber ); ...
                           zeros( ( nfft / 2 + 1 ) - idxHighBP - 1, refMicNumber ) ];
        freqWindowing  = single( freqWindowing );
        freqWindowing  = repelem( freqWindowing, 1, 1, virtualMicNumber );
        freqWindowing  = permute( freqWindowing, [ 1 3 2 ] );
             
    end  
end

fftNormFactor           = nfft;  % define normalization factor of FFT. 

if coherenceEnable
    
    % define parameters for overlapped FFT
    win                 = single( hann( nfft ) );
    energyNormFactor    = ( rms( win ) ) ^-1;
    if mimoEnable
        win             = repelem( win, 1, refMicNumber );
    else
        win             = repelem( win, 1, virtualMicNumber );
    end
    fftStep             = nfft/2; % overlap is 50%
    fftStepNumber       = ceil( bufferSize / fftStep ) - 1;
    if fftStepNumber == 0
        fftStepNumber = 1;
        fftNormFactor = sqrt( bufferSize * nfft ); % correct normalization factor
    end
    
    % calculate total TF if needed
    if totalTFenable
        
        if mimoEnable
            
            fprintf( '\nCalculating total MIMO transfer function\n\n' );
            transferFunction    = tfestimate( referenceMic, spsFormat, hann( nfft ), nfft/4, nfft, 'mimo' );
            
        else
            
            fprintf( '\nCalculating total transfer function\n\n' );
            transferFunction    = tfestimate( referenceMic, spsFormat, hann( nfft ), nfft/4, nfft );

        end
        
        % perform filtering of TF if needed
        if filteringEnable
                        
            transferFunction    = transferFunction .* freqWindowing;
            
        end
        
    end
    
end

fprintf( [ 'Processing ' num2str( numbersFrame ) ' frames.\n' ] );
splMatrix               = zeros ( surfpoint/2, surfpoint, numbersFrame );   % preallocation

% Percentage initializing
oldCompletedPercentage  = -1;
fprintf( '\nFrames processing: ' )
varStr                  = '0%%';
varStrLen               = length( varStr );
fprintf( '%s', varStr )

for blockIndex = 1 : numbersFrame
    
    %%%%%%%%%%% PERCENTAGE PROCESSING %%%%%%%%%%%  
    completedPercentage     = round( blockIndex/numbersFrame*100 );
    if oldCompletedPercentage ~= completedPercentage
        for i = 1 : varStrLen
            fprintf( '\b' )
        end
        varStr              = sprintf( '%d%%', completedPercentage );
        varStrLen           = length( varStr );
        fprintf( '%s', varStr )
    end
    oldCompletedPercentage  = completedPercentage;
    
    [ ZI ] = frameProcessing( blockIndex, coherenceEnable, referenceMic, stepSize, bufferSize, ...
            totalTFenable, mimoEnable, spsFormat, nfft, filteringEnable, freqWindowing, refMicNumber, ...
            fftStepNumber, fftStep, win, energyNormFactor, virtualMicNumber, fullScaleLevel, ...
            extX, extY, extXI, extYI, surfpoint, transferFunction, fftNormFactor );
            
    splMatrix( :, :, blockIndex )   = ZI;

end

splMatrix           = single( splMatrix );

% pointing correction
correctionIndex     = round( horizontalTilt * surfpoint / 360 );
if correctionIndex > 0  
    splMatrix       = splMatrix( :, [ correctionIndex+1 : surfpoint, 1:correctionIndex ], : );
elseif correctionIndex < 0
    splMatrix       = splMatrix( :, [ surfpoint+correctionIndex+1 : surfpoint, 1:surfpoint+correctionIndex ], : );
end

upperLimit          = max( max( max( splMatrix ) ) );                       % find absolute max level
lowerLimit          = upperLimit - noiseGate;                               % range to plot

if useFixedSplPlotBounds == 1
    v               = fixedLowerLimit : contourSpan : fixedUpperLimit;      % iso-line levels
    splMatrix( splMatrix < fixedLowerLimit ) = 0;                           % set to 0 SPL values < lowerLimit
else                         
    v               = lowerLimit+1 : contourSpan : upperLimit;              % iso-line levels
    splMatrix( splMatrix < lowerLimit ) = 0;                                % set to 0 SPL values < lowerLimit
end


if timeSignalEnable
    spsMean         = mean( spsFormat, 2 );
end

fprintf( '\nData processing ended correctly.\n' )


%% VIDEO PROCESSING
fprintf( '\nInitializing video processing.\n' );
if coherenceEnable
    referenceName   = deblank( referenceName );
    videoTitle      = [ recordedFileName( 1:end-4 ) '_' referenceName ];    % name of file video
else
    videoTitle      = recordedFileName( 1 : end-4 );                        % name of file video
end
videoFile           = VideoWriter( videoTitle, 'MPEG-4' );                  % initializes video
videoFile.FrameRate = frameRate;                                            % set frame rate
videoFile.Quality   = videoQuality;                                         % set quality
open( videoFile );
cmap                = jet( 64 );
cmap( 1, : )        = [ 0 0 0 ]; % black full scale

% Percentage initializing
oldCompletedPercentage  = -1;
fprintf( '\nVideo processing: ' )
varStr                  = '0%%';
varStrLen               = length( varStr );
fprintf( '%s', varStr )

if timeSignalEnable

    verticalLimit   = [ min( spsMean ) max( spsMean ) ];
    videoHeigth     = mapWindowWidth / ( 2 * 0.8 );                      % define video window
    videoWindow     = figure( 'units', 'pixel', 'innerposition', [ 0 0 mapWindowWidth videoHeigth  ], 'menubar', 'none' );

    for blockIndex = 1 : numbersFrame
        
        %%%%%%%%%%% PERCENTAGE PROCESSING %%%%%%%%%%%  
        completedPercentage     = round( blockIndex/numbersFrame*100 );
        if oldCompletedPercentage ~= completedPercentage
            for i = 1 : varStrLen
                fprintf( '\b' )
            end
            varStr              = sprintf( '%d%%', completedPercentage );
            varStrLen           = length( varStr );
            fprintf( '%s', varStr )
        end
        oldCompletedPercentage  = completedPercentage;
        
        ZI          = splMatrix( :, :, blockIndex );                        % extract current value to plot

        % subplot colormap
        plotBG      = subplot( 'Position', [ 0 0.2 1 0.8 ] );
        videoColorMap( background, backgroundWidth, backgroundHeight, XI, YI, ZI, lowerLimit, upperLimit, plotBG, ...
                       v, alpha, contourEnable, useFixedSplPlotBounds, fixedLowerLimit, fixedUpperLimit, valueTickEnable, cmap );
        
        % subplot time signal
        plotTime   	= subplot( 'Position', [ 0 0 1 0.2 ] );                 % subplot with time signal
        timeSignalPlot( spsMean, verticalLimit, blockIndex, stepSize, bufferSize );

        writeVideo( videoFile, getframe( videoWindow ) );
        
    end
    
elseif outputWidth ~= 0
    
    videoHeigth     = mapWindowWidth / 2;                                % set video window dimension to 2:1 for VR
    videoWindow   	= figure( 'units', 'pixel', 'innerposition', [ 0 0 mapWindowWidth videoHeigth  ], 'menubar', 'none' );

    for blockIndex = 1 : numbersFrame

        %%%%%%%%%%% PERCENTAGE PROCESSING %%%%%%%%%%%  
        completedPercentage     = round( blockIndex/numbersFrame*100 );
        if oldCompletedPercentage ~= completedPercentage
            for i = 1 : varStrLen
                fprintf( '\b' )
            end
            varStr              = sprintf( '%d%%', completedPercentage );
            varStrLen           = length( varStr );
            fprintf( '%s', varStr )
        end
        oldCompletedPercentage  = completedPercentage;

        ZI              = splMatrix(:,:,blockIndex);                    % extract current value to plot

        plotBG          = subplot( 'Position', [ 0 0 1 1 ] );
        set( plotBG, 'YDir', 'normal' );
        videoColorMapFFMPEG( XI, YI, ZI, lowerLimit, upperLimit, v, contourEnable, ...
                             useFixedSplPlotBounds, fixedLowerLimit, fixedUpperLimit, valueTickEnable, cmap );

        writeVideo( videoFile, getframe( videoWindow ) );
        
    end
    
    close( videoFile )
    % copy background image to root folder and convert to requested format
    copyfile( backgroundImageFile, './', 'f' )

    if outputBlackWhite
        system( [ 'ffmpeg -y -i ' backgroundImageFileName ' -vf hue=s=0 ' backgroundImageFileName ] )
    end
    
    system( [ 'ffmpeg -i ' backgroundImageFileName ' -vf scale=' num2str( outputWidth ) ':' num2str( outputWidth/2 ) ' temp.png ' ] );
    delete( backgroundImageFileName )
    
    % create background video from image
    mapVideoFile            = [ videoTitle '.mp4' ];
    backgroundVideoOutput   = [ 'overlay_' mapVideoFile ];
    duration                = stepSize * numbersFrame / fs;
    system( [ 'ffmpeg -loop 1 -i temp.png -c:v libx264 -t ' num2str( duration ) ' -pix_fmt yuv420p tempBackground.mp4' ] );
    delete( 'temp.png' )
    clear mapVideo 
    
    % generate overlay
    system( [ 'ffmpeg -y -i ' mapVideoFile ' -i tempBackground.mp4 -filter_complex " '...
    '[0:v]setpts=PTS-STARTPTS, scale=' num2str( outputWidth ) 'x' num2str( outputWidth/2 ) '[top]; '...
    '[1:v]setpts=PTS-STARTPTS, scale=' num2str( outputWidth ) 'x' num2str( outputWidth/2 ) '[bottom]; '...
    '[top]split[m][a]; [m][a]alphamerge[keyed]; [bottom][keyed]overlay=eof_action=endall" ' backgroundVideoOutput ] );        
    delete( 'tempBackground.mp4' )
    delete( mapVideoFile )
    movefile( backgroundVideoOutput, mapVideoFile );
        
else
    
    videoHeigth     = mapWindowWidth / 2;                                % set video window dimension to 2:1 for VR
    videoWindow   	= figure( 'units', 'pixel', 'innerposition', [ 0 0 mapWindowWidth videoHeigth  ], 'menubar', 'none' );

    for blockIndex = 1 : numbersFrame
        
        %%%%%%%%%%% PERCENTAGE PROCESSING %%%%%%%%%%%  
        completedPercentage     = round( blockIndex/numbersFrame*100 );
        if oldCompletedPercentage ~= completedPercentage
            for i = 1 : varStrLen
                fprintf( '\b' )
            end
            varStr              = sprintf( '%d%%', completedPercentage );
            varStrLen           = length( varStr );
            fprintf( '%s', varStr )
        end
        oldCompletedPercentage  = completedPercentage;
        
        ZI          = splMatrix(:,:,blockIndex);                            % extract current value to plot

        % subplot colormap
        plotBG      = subplot( 'Position', [ 0 0 1 1 ] );
        videoColorMap( background, backgroundWidth, backgroundHeight, XI, YI, ZI, lowerLimit, upperLimit, plotBG, ...
                       v, alpha, contourEnable, useFixedSplPlotBounds, fixedLowerLimit, fixedUpperLimit, valueTickEnable, cmap );
                  
        writeVideo( videoFile, getframe( videoWindow ) );
        
    end
end

close( videoFile );                                                         % finalizes video 
close all


%% AUDIO PROCESSING
if audioMono == 1 || ambix_o1 == 1 || ambix_o3 == 1
    
    % Load ambix matrix and EM recorded for audio of VR video ( if selected )
    if audioFromEm32 % load ambix matrix and EM32recorded file
        
        ambixMatrix         = [ resourcesFolder ambixMatrixFileName ];
        disp( [ 'Loading ambix matrix 16ch 3rd order: ' ambixMatrix ] );
        [ ambixIR, fs ]     = load_ir( ambixMatrix, 32, 16 );                   % ambix 3rd order is [32-IN x 16-OUT x Nsamples]

        disp( [ 'Loading Eigenmike 32 channels recording: ' recordedFileName ] );
        [ EMrecorded, fs ]  = audioread( recordedNoiseFile );
        
        EMrecorded          = EMrecorded( 1 + noiseStart : noiseEnd, : );
        % apply trim and zeropad matrix
        EMrecorded          = [ zeroPadMatrix( :, 1:32 ); EMrecorded; zeroPadMatrix( :, 1:32 ) ];

        if audioCoherence
            
            if totalTFenable
        
                fprintf( '\nFiltering EM recorded sound with static Transfer Function\n' )
                [ EMrecorded ]  = staticTFfilter( referenceMicAudio, EMrecorded, bandpassLowCutFreq, bandpassHighCutFreq, filteringEnable, fs );
        
            else  

                fprintf( '\nFiltering EM recorded sound with running Transfer Function\n' )
                [ EMrecorded ]  = movingTFfilter( referenceMicAudio, EMrecorded, bandpassLowCutFreq, bandpassHighCutFreq, filteringEnable, fs );
                    
            end
        
        end
        
        % AMBIX CONVOLUTION
        fprintf( '\nConvolving EM32 with Ambix FIR matrix.\n' );
        ambixAudio_o3           = ambixConv( EMrecorded, ambixIR );
        clear ambixIR ambixMatrix EMrecorded
        
        if audioCoherence % correct orientation of sound field

            % FLIP: invert right left
            ambixAudio_o3(:,1)  = -ambixAudio_o3(:,1);
            ambixAudio_o3(:,4)  = -ambixAudio_o3(:,4);
            ambixAudio_o3(:,5)  = -ambixAudio_o3(:,5);
            ambixAudio_o3(:,9)  = -ambixAudio_o3(:,9);
            ambixAudio_o3(:,10) = -ambixAudio_o3(:,10);
            ambixAudio_o3(:,11) = -ambixAudio_o3(:,11);
            % FLOP: invert front back
            ambixAudio_o3(:,3)  = -ambixAudio_o3(:,3);
            ambixAudio_o3(:,4)  = -ambixAudio_o3(:,4);
            ambixAudio_o3(:,7)  = -ambixAudio_o3(:,7);
            ambixAudio_o3(:,10) = -ambixAudio_o3(:,10);
            ambixAudio_o3(:,13) = -ambixAudio_o3(:,13);
            ambixAudio_o3(:,15) = -ambixAudio_o3(:,15);

        end
    
    elseif audioFromSps
        
        fprintf( '\nAudio will be calculated from SPS format\n' )
        fprintf( '\nSelect virtual mics\n' )
        backgroundFrame         = readFrame( backgroundVideo, 'native' );               % extract background frame
        [ backgroundFrame, imageHeight, imageWidth ] = image2x1resize( backgroundFrame, mapWindowWidth ); % adjust resolution
        [ spsFormatAudio, azimuth, elevation ] = selectSPSsource( backgroundFrame, X, Y, spsAudio );

        if audioCoherence
        
            if totalTFenable
        
                fprintf( '\nFiltering virtual mics with static Transfer Function\n' )
                [ spsFormatAudio ] = staticTFfilter1to1( referenceMicAudio, spsFormatAudio, bandpassLowCutFreq, bandpassHighCutFreq, filteringEnable );

            else  

                fprintf( '\nFiltering virtual mics with running Transfer Function\n' )
                [ spsFormatAudio ] = movingTFfilter1to1( referenceMicAudio, spsFormatAudio, bandpassLowCutFreq, bandpassHighCutFreq, filteringEnable );

            end
            
        end
        
        fprintf( '\nPanning sources in Ambisonics format\n' )
        [ ambixAudio_o3 ]   = panAmbix( azimuth, elevation, spsFormatAudio );
        
    elseif audioFromReference     
    
        fprintf( '\nAutopanning audio with coherence analysis between reference ans SPS format\n' )
        [ ambixAudio_o3 ]   = autoPan( spsAudio, referenceMicAudio );
        
    end
    
    if filteringAudio && audioCoherence == 0 % with coherence, filtering is performed on transfer function 
    
        fprintf( 'Performing ambix audio filtering.\n' );

        d = designfilt( 'bandpassfir', ...              % response type
                        'FilterOrder', 20, ...          % Filter order
                        'StopbandFrequency1',   0.8 * bandpassLowCutFreq,   ... % Frequency constraint
                        'PassbandFrequency1',   bandpassLowCutFreq,         ... % Frequency constraint
                        'PassbandFrequency2',   bandpassHighCutFreq,        ... % Frequency constraint
                        'StopbandFrequency2',   1.2 * bandpassHighCutFreq,  ... % Frequency constraint
                        'DesignMethod', 'ls', ...        % Design method
                        'StopbandWeight1', 3, ... 
                        'PassbandWeight', 1, ...
                        'StopbandWeight2', 3, ...
                        'SampleRate',           fs );

        ambixAudio_o3       = filter( d, ambixAudio_o3 );                   % perform filtering

    end
    
    % CORRECT ALIGNMENT
    ambixAudio_o3 = [ ambixAudio_o3( bufferSize+1 : end, : ); zeros( bufferSize, 16 ) ];

    % NORMALIZE AND SAVE
    if audioMono % extract, normalize and save
        omniAudio           = ambixAudio_o3( :, 1 );                        % keep only ambix omni mic
        maxAbsValue         = max( max( abs( omniAudio ) ) );               % search for absolute max value
        omniAudio           = omniAudio ./ maxAbsValue;                     % normalize between -1 and +1
        outputWavOmni       = [ recordedFileName( 1:end-4 ) '_omni.wav' ];
        fprintf( [ '\nSaving omni audio format in WAV file: ' outputWavOmni ] );
        audiowrite( outputWavOmni, omniAudio, fs, 'BitsPerSample', audioBitsPerSample );
        fprintf( '\nOmni audio saved correctly.\n' );
    end

    if ambix_o1
        outputWavAmbix_o1   = [ recordedFileName( 1:end-4 ) '_ambix_o1.wav' ];  % define filename
        ambixAudio_o1       = ambixAudio_o3( :, 1:4 );                          % keep only first four channels
        maxAbsValue         = max( max( abs( ambixAudio_o1 ) ) );               % search for absolute max value
        ambixAudio_o1       = ambixAudio_o1 ./ maxAbsValue;                     % normalize between -1 and +1
        fprintf( [ '\nSaving 1st order Ambix format in WAV file: ' outputWavAmbix_o1 ] );
        audiowrite( outputWavAmbix_o1, ambixAudio_o1, fs, 'BitsPerSample', audioBitsPerSample );   % save file
        fprintf( '\nAmbix format saved correctly.\n' );
    end
    
    if ambix_o3
        outputWavAmbix_o3      = [ recordedFileName( 1:end-4 ) '_ambix_o3.wav' ];  % define filename
        maxAbsValue         = max( max( abs( ambixAudio_o3 ) ) );               % search for absolute max value
        ambixAudio_o3       = ambixAudio_o3 ./ maxAbsValue;                     % normalize between -1 and +1
        fprintf( [ '\nSaving 3rd order Ambix format in WAV file: ' outputWavAmbix_o3 ] );
        audiowrite( outputWavAmbix_o3, ambixAudio_o3, fs, 'BitsPerSample', audioBitsPerSample );   % save file
        fprintf( '\nAmbix format saved correctly.\n' );
    end
       
end

%% ADD AUDIO TO OVERLAY
videoFileName           = [ videoTitle '.mp4' ];

if audioMono
	fprintf( '\nAdding omni audio\n' )
    outputFileName      = [ videoTitle '_omni.mp4' ];
    system( [ 'FFMPEG -i ' videoFileName ' -i ' outputWavOmni ' -map 0:v -map 1:a -c:v copy -shortest ' outputFileName ] );
    delete( outputWavOmni )
end

if ambix_o1
    fprintf( '\nAdding 1st order Ambix audio\n' )
    outputFileName       = [ videoTitle '_ambix_o1.mov' ];
    system( [ 'FFMPEG -i ' videoFileName ' -i ' outputWavAmbix_o1 ' -map 0:v -map 1:a -c:v copy -c:a copy -shortest ' outputFileName ] );
    delete( outputWavAmbix_o1 )
end

if ambix_o3
    fprintf( '\nAdding 3rd order Ambix audio\n' )
    outputFileName       = [ videoTitle '_ambix_o3.mov' ];
    system( [ 'FFMPEG -i ' videoFileName ' -i ' outputWavAmbix_o3 ' -map 0:v -map 1:a -c:v copy -c:a copy -shortest ' outputFileName ] );
    delete( outputWavAmbix_o3 )
end

if audioMono || ambix_o1 || ambix_o3
    delete( videoFileName )
end

if injectMetadata
    source              = './lib/SpatialMediaMetadataInjector.exe';
    copyfile( source, './', 'f' )
    system( 'SpatialMediaMetadataInjector.exe' )
end

delete( 'SpatialMediaMetadataInjector.exe' )
rmpath( './lib' )
fprintf( '\nEnded correctly.\n\n' )
