% autoPan.m This function resolve Ambisonics equation (depending on
% azimuth and elevation) and collocate virtual sources in space.

function [ ambixAudio ] = autoPan( spsFormat, referenceMic )

positionBuffer          = 10; % even number >= 2
referenceNum            = size( referenceMic, 2 );
virtualMicNum           = size( spsFormat, 2 );
spsLength               = size( spsFormat, 1 );
               

%% DEFINE NFFT and BUFFERSIZE
fprintf( 'Define nfft and bufferSize values for audio coherence analysis\n' )
nfft                    = input( 'nfft: ' );
bufferSize              = input( 'bufferSize: ' );

spsFormat               = [ spsFormat; zeros( bufferSize, virtualMicNum ) ];
referenceMic            = [ referenceMic; zeros( bufferSize, referenceNum ) ];

chunkNumber             = floor( length( spsFormat ) / bufferSize );
noiseEnd                = bufferSize * chunkNumber;

spsFormat               = spsFormat( 1 : noiseEnd, : );
referenceMic            = referenceMic( 1 : noiseEnd, : );


%% MAX COHERENCE OVER TIME
fprintf( '\nCalculating running max coherence\n' );
pointsRadiant           = GenerateUniformTestDir( 4 ); % generation of mic coordinates
pointsRadiant( :, 2 )   = -pointsRadiant( :, 2 ) .* cos( pointsRadiant( :, 2 ) ./ pi/2 );

for j = 1 : referenceNum
    
    for i = 1 : chunkNumber

        refBlock            = referenceMic( (i-1)*bufferSize + 1 : (i-1)*bufferSize + bufferSize, j );
        spsBlock            = spsFormat( (i-1)*bufferSize + 1 : (i-1)*bufferSize + bufferSize, : );

        [ coherence ]       = mscohere( spsBlock, refBlock, hann( nfft ), nfft/4, nfft ); 

        [ ~, maxCoh ]       = max( mean( coherence, 1 ) );

        position(i,:,j)     = pointsRadiant( maxCoh,: );            

    end

end

% correction of position
for j = 1 : referenceNum
    
    for i = 2 : chunkNumber -1

        if position(i-1,1,j) == position(i+1,1,j)
            position(i,1,j) = position(i-1,1,j);
        end

        if position(i-1,2,j) == position(i+1,2,j)
            position(i,2,j) = position(i-1,2,j);
        end

    end

end

% averaging position
for j = 1 : referenceNum
    
    for i = 1 : chunkNumber - positionBuffer

        buffer              = position( i : i + positionBuffer-1, :, j );
        positionCorrected(i,:,j) = mean( buffer, 1 ); 

    end

end

firstPos            = positionCorrected( 1, :, : );
firstPosRep         = repelem( firstPos, positionBuffer/2, [1] );
lastPos             = positionCorrected( end, :, : );
lastPosRep          = repelem( lastPos, positionBuffer/2, [1] );
positionCorrected   = [ firstPosRep; positionCorrected; lastPosRep ];


%% MOVING PAN
fprintf( '\nGenerating moving signals in ambix format\n' );
audioPanned     = zeros( 1+length( referenceMic ), 16, referenceNum );

for i = 1 : referenceNum
    
    refPan      = zeros( 1, 16 );
    
    for j = 1 : chunkNumber

        refBlock = referenceMic( (j-1)*bufferSize + 1 : (j-1)*bufferSize + bufferSize, i );

        a       = positionCorrected( j, 1, i );
        e       = positionCorrected( j, 2, i );

        %% AMBIX FORMAT
        % first order
        A(1,1)  = 1;                                                            % W
        A(1,2)  = sin(a)*cos(e);                                                % Y
        A(1,3)  = sin(e);                                                       % Z
        A(1,4)  = cos(a)*cos(e);                                                % X
        % second order         
        A(1,5)  = sqrt(3/4)*sin(2*a)*cos(e)*cos(e);                             % R
        A(1,6)  = sqrt(3/4)*sin(a)*sin(2*e);                                    % S
        A(1,7)  = (1/2)*(3*sin(e)*sin(e)-1);                                    % T
        A(1,8)  = sqrt(3/4)*cos(a)*sin(2*e);                                    % U
        A(1,9)  = sqrt(3/4)*cos(2*a)*cos(e)*cos(e);                             % V
        % third order       
        A(1,10) = sqrt(5/8)*sin(3*a)*cos(e)*cos(e)*cos(e);                      % K
        A(1,11) = sqrt(15/4)*sin(2*a)*sin(e)*cos(e)*cos(e);                     % L
        A(1,12) = sqrt(3/8)*sin(a)*cos(e)*(5*sin(e)*sin(e)-1);                  % M
        A(1,13) = (1/2)*sin(e)*(5*sin(e)*sin(e)-3);                             % N
        A(1,14) = sqrt(3/8)*cos(a)*cos(e)*(5*sin(e)*sin(e)-1);                  % O
        A(1,15) = sqrt(15/4)*cos(2*a)*sin(e)*cos(e)*cos(e);                     % P
        A(1,16) = sqrt(5/8)*cos(3*a)*cos(e)*cos(e)*cos(e);                      % Q

        refPan  = [ refPan; refBlock .* A ];

    end
    
    audioPanned( :, :, i ) = refPan;
    
end

audioPanned     = audioPanned( 2 : spsLength+1, :, : );
audioPanned     = sum( audioPanned, 3 );
maxAbsValue     = max( max( abs( audioPanned ) ) );                         % search for absolute max value
ambixAudio      = audioPanned ./ maxAbsValue;
