% movingTFfilter.m This function perform overlap and add convolution with analysis on em32recorded
% tracks using a moving transfer function

function [ audioCoherent ]    = movingTFfilter( reference, signal, bandpassLowCutFreq, bandpassHighCutFreq, ...
                                                filteringEnable, fs )

                        
virtualMicNumber        = size( signal, 2 );
referenceNumber         = size( reference, 2 );

if filteringEnable
    nfftMin             = 2 ^ ( nextpow2( 2 * ( bandpassHighCutFreq - bandpassLowCutFreq ) ) );
    fprintf( [ 'Nfft minimum value for correct filtering is: ' num2str( nfftMin ) '\n' ] )
else
    nfftMin             = 0;
end

% define nfft value
[ nfft ]                = checkTFnfft2( reference, signal );
if nfft < nfftMin
    fprintf( [ '\nChosed value ( ' num2str( nfft ) ' ) is not allowed\n' ] )
    fprintf( [ 'Nfft setted to min possible value ( ' num2str( nfftMin ) ' )\n' ] )
    nfft                = nfftMin;
    fprintf( 'Press any key to continue or CTRL+C to abort\n' )
    pause
end

% define bufferSize for moving transfer function
prompt                  = 'Define bufferSize for audio processing: ';
N                       = input( prompt );

% define filtering windows
if filteringEnable
    
    [ freqWindowing ]   = bandpassFreqFilt( bandpassHighCutFreq, bandpassLowCutFreq, ...
                                            nfft, fs, virtualMicNumber ); % generate freq. window   
                                                   
    freqWindowing       = single( freqWindowing ); % convert data type

end

% define transfer function window
tfWin                   = repelem( hann ( nfft ), 1, virtualMicNumber );

% preallocate memory
audioCoherent           = zeros( size( reference', 2 ), virtualMicNumber, referenceNumber );
audioCoherent           = single( audioCoherent );

for j = 1 : referenceNumber
    
    fprintf( [ 'Filtering with reference ' num2str( j ) '\n' ] )
    
    %% OA MULTI CHANNEL CONV %%%%%%%%%%
    Nin                 = size( reference', 2 );
    M                   = 1;
    V                   = virtualMicNumber;

    BlocksNum           = ceil(Nin/N);
    BlockId             = 1;

    inData              = [ reference(:,j)', zeros( 1, BlocksNum*N-Nin ) ];         % zeropad reference
    refData             = [ signal', zeros( virtualMicNumber, BlocksNum*N-Nin ) ];  % zeropad em32rec
    Lim                 = length( inData );

    outData             = zeros( V, BlocksNum*N+N );
    overlapBlock        = zeros( V, N );
    fd_convBlock        = zeros( 1, V, N );

    % percentage initializing
    oldCompletedPercentage = -1;
    fprintf( '\nConvolving audio: ' )
    varStr              = '0%%';
    varStrLen           = length( varStr );
    fprintf( '%s', varStr )

    % Process the data block by block 
    while( BlockId <= BlocksNum )

        %%%%%%%%%%% PERCENTAGE PROCESSING %%%%%%%%%%%

        CompletedPercentage     = round(BlockId/BlocksNum*100);
        if oldCompletedPercentage ~= CompletedPercentage
            for i = 1 : varStrLen
                fprintf( '\b' )
            end
            varStr      = sprintf( '%d%%', CompletedPercentage );
            varStrLen   = length( varStr );
            fprintf( '%s', varStr )
        end
        oldCompletedPercentage  = CompletedPercentage;

        %%%%%%%%%%% SIGNAL PROCESSING %%%%%%%%%%%

        % inBlock size [Mic x N]
        inBlock         = inData( :, ( BlockId-1 )*N+1 : min( BlockId*N, Lim ) );    % reference
        refBlock        = refData( :, ( BlockId-1 )*N+1 : min( BlockId*N, Lim ) );   % em32rec

        % compute transfer function
        transFunct      = tfestimate( inBlock', refBlock', hann( nfft ), nfft/2, nfft );

        if filteringEnable

            transFunct  = transFunct .* freqWindowing; % apply filtering window

        end

        % back in time domain
        transFunct      = [ zeros(1,virtualMicNumber); transFunct( 2 : end - 1,: ); ...
                            zeros(1,virtualMicNumber); flipud( conj( transFunct( 2 : end - 1, : ) ) ) ];
        tfIR            = fftshift( ifft( transFunct ) );
        tfIR            = tfIR .* tfWin;

        [ ~, delay ]    = max( tfIR( :, 1 ) );
        allDelay( BlockId ) = delay;
        h               = ones( 1, virtualMicNumber, nfft );
        h( 1, :, : )    = tfIR';
        H               = fft( h, 2*N, 3 );

        %%%%%%%%%%% BLOCK PROCESSING %%%%%%%%%%%

        fd_inBlock      = fft( inBlock, 2*N, 2 );

        for k = 1:1:2*N
            fd_convBlock( 1, :, k ) = reshape( fd_inBlock(:,k), 1, M ) * squeeze( H( :, :, k ) );
        end

        convBlock       = ifft( reshape( fd_convBlock, V, 2*N ), 2*N, 2 );
        outBlock        = overlapBlock + convBlock(:,1:N);
        overlapBlock    = [ convBlock(:,N+1:2*N-1) , zeros(V,1) ];
        outData(:,(BlockId-1)*N+1:BlockId*N) = outBlock;

        BlockId         = BlockId+1;

    end
    fprintf( '\n' )
    outData( :, (BlockId-1)*N+1:BlockId*N ) = overlapBlock; % Store last overlapBlock

    % calculate mean delay
    allDelay            = allDelay( 2 : end-1 );
    delayMean           = mean( allDelay );

    % crop outdata
    outData             = outData( :, delayMean : Nin+N-1 );
    outData             = outData';
    outData             = outData( 1 : length( signal ), : );
    outData( isnan( outData ) ) = 0;
    outData             = single( outData );
    
    audioCoherent( :, :, j ) = outData;
    
end

fprintf ( '\nDONE\n' )

audioCoherent           = sum( audioCoherent, 3 );
